# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Common functionality for the transformations/optimization pipeline."""

from typing import Any, Container, Optional, Sequence, TypeVar, Union

import dace
from dace import data as dace_data, subsets as dace_sbs, symbolic as dace_sym
from dace.sdfg import graph as dace_graph, nodes as dace_nodes
from dace.transformation import pass_pipeline as dace_ppl
from dace.transformation.passes import analysis as dace_analysis

from gt4py.next.program_processors.runners.dace import utils as gtx_dace_utils


_PassT = TypeVar("_PassT", bound=dace_ppl.Pass)


def gt_make_transients_persistent(
    sdfg: dace.SDFG,
    device: dace.DeviceType,
) -> dict[int, set[str]]:
    """
    Changes the lifetime of certain transients to `Persistent`.

    A persistent lifetime means that the transient is allocated only the very first
    time the SDFG is executed and only deallocated if the underlying `CompiledSDFG`
    object goes out of scope. The main advantage is, that memory must not be
    allocated every time the SDFG is run. The downside is that the SDFG can not be
    called by different threads.

    Args:
        sdfg: The SDFG to process.
        device: The device type.

    Returns:
        A `dict` mapping SDFG IDs to a set of transient arrays that
        were made persistent.

    Note:
        This function is based on a similar function in DaCe. However, the DaCe
        function does, for unknown reasons, also reset the `wcr_nonatomic` property,
        but only for GPU.
    """
    result: dict[int, set[str]] = {}
    for nsdfg in sdfg.all_sdfgs_recursive():
        fsyms: set[str] = nsdfg.free_symbols
        modify_lifetime: set[str] = set()
        not_modify_lifetime: set[str] = set()

        for state in nsdfg.states():
            scope_dict = state.scope_dict()
            for dnode in state.data_nodes():
                if dnode.data in not_modify_lifetime:
                    continue

                if dnode.data in nsdfg.constants_prop:
                    not_modify_lifetime.add(dnode.data)
                    continue

                desc = dnode.desc(nsdfg)
                if not desc.transient or type(desc) not in {dace.data.Array, dace.data.Scalar}:
                    not_modify_lifetime.add(dnode.data)
                    continue
                if desc.storage == dace.StorageType.Register:
                    not_modify_lifetime.add(dnode.data)
                    continue

                if desc.lifetime == dace.AllocationLifetime.External:
                    not_modify_lifetime.add(dnode.data)
                    continue

                # If the data is referenced inside a scope, such as a map, it might be possible
                #  that it is only used inside that scope. If we would make it persistent, then
                #  it would essentially be allocated outside and be shared among the different
                #  map iterations. So we can not make it persistent.
                #  The downside is, that we might have to perform dynamic allocation.
                if scope_dict[dnode] is not None:
                    not_modify_lifetime.add(dnode.data)
                    continue

                try:
                    # The symbols describing the total size must be a subset of the
                    #  free symbols of the SDFG (symbols passed as argument).
                    # NOTE: This ignores the renaming of symbols through the
                    #   `symbol_mapping` property of nested SDFGs.
                    if not set(map(str, desc.total_size.free_symbols)).issubset(fsyms):
                        not_modify_lifetime.add(dnode.data)
                        continue
                except AttributeError:  # total_size is an integer / has no free symbols
                    pass

                # Make it persistent.
                modify_lifetime.add(dnode.data)

        # Now setting the lifetime.
        result[nsdfg.cfg_id] = modify_lifetime - not_modify_lifetime
        for aname in result[nsdfg.cfg_id]:
            nsdfg.arrays[aname].lifetime = dace.AllocationLifetime.Persistent

    return result


def gt_find_constant_arguments(
    call_args: dict[str, Any],
    include: Optional[Container[str]] = None,
) -> dict[str, Any]:
    """Scans the calling arguments for compile time constants.

    The output of this function can be used as input to
    `gt_substitute_compiletime_symbols()`, which then removes these symbols.

    By specifying `include` it is possible to force the function to include
    additional arguments, that would not be matched otherwise. Importantly,
    their value is not checked.

    Args:
        call_args: The full list of arguments that will be passed to the SDFG.
        include: List of arguments that should be included.
    """
    if include is None:
        include = set()
    ret_value: dict[str, Any] = {}

    for name, value in call_args.items():
        if name in include or (gtx_dace_utils.is_stride_symbol(name) and value == 1):
            ret_value[name] = value

    return ret_value


def is_accessed_downstream(
    start_state: dace.SDFGState,
    sdfg: dace.SDFG,
    data_to_look: str,
    reachable_states: Optional[dict[dace.SDFGState, set[dace.SDFGState]]],
    nodes_to_ignore: Optional[set[dace_nodes.AccessNode]] = None,
    states_to_ignore: Optional[set[dace.SDFGState]] = None,
) -> bool:
    """Scans for accesses to the data container `data_to_look`.

    The function will go through states that are reachable from `start_state`
    (included) and test if there is an AccessNode that _reads_ from `data_to_look`.
    It will return `True` the first time it finds such a node.

    The function will ignore all nodes that are listed in `nodes_to_ignore`.
    Furthermore, states listed in `states_to_ignore` will be ignored, i.e.
    handled as they did not exist.

    Args:
        start_state: The state where the scanning starts.
        sdfg: The SDFG on which we operate.
        data_to_look: The data that we want to look for.
        reachable_states: Maps an `SDFGState` to all `SDFGState`s that can be reached.
            If `None` it will be computed, but this is not recommended.
        nodes_to_ignore: Ignore these nodes.
        states_to_ignore: Ignore these states.

    Note:
        Currently, the function will not only ignore the states that are listed in
        `states_to_ignore`, but all that are reachable from any of these states.
        Thus care must be taken when this option is used. Furthermore, this behaviour
        is not intended and will change in further versions.
        `reachable_states` can be computed by using the `StateReachability` analysis
        pass from DaCe.

    Todo:
        - Modify the function such that it is no longer necessary to pass the
            `reachable_states` argument.
        - Fix the behaviour for `states_to_ignore`.
    """
    # After DaCe 1 switched to a hierarchical version of the state machine. Thus
    #  it is no longer possible in a simple way to traverse the SDFG. As a temporary
    #  solution we use the `StateReachability` pass. However, this has some issues,
    #  see the note about `states_to_ignore`.
    if reachable_states is None:
        state_reachability_pass = dace_analysis.StateReachability()
        reachable_states = state_reachability_pass.apply_pass(sdfg, None)[sdfg.cfg_id]
    else:
        # Ensures that the externally generated result was passed properly.
        assert all(
            isinstance(state, dace.SDFGState) and state.sdfg is sdfg for state in reachable_states
        )

    ign_dnodes: set[dace_nodes.AccessNode] = nodes_to_ignore or set()
    ign_states: set[dace.SDFGState] = states_to_ignore or set()

    # NOTE: We have to include `start_state`, however, we must also consider the
    #  data in `reachable_states` as immutable, so we have to do it this way.
    # TODO(phimuell): Go back to a trivial scan of the graph.
    if start_state not in reachable_states:
        # This can mean different things, either there was only one state to begin
        #  with or `start_state` is the last one. In this case the `states_to_scan`
        #  set consists only of the `start_state` because we have to process it.
        states_to_scan = {start_state}
    else:
        # Ensure that `start_state` is scanned.
        states_to_scan = reachable_states[start_state].union([start_state])

    # In the first version we explored the state machine and if we encountered a
    #  state in the ignore set we simply ignored it. This is no longer possible.
    #  Instead we will remove all states from the `states_to_scan` that are reachable
    #  from an ignored state. However, this is not the same as if we would explore
    #  the state machine (as we did before). Consider the following case:
    #
    #   (STATE_1) ------------> (STATE_2)
    #       |                       /\
    #       V                       |
    #   (STATE_3) ------------------+
    #
    #  Assume that `STATE_1` is the starting state and `STATE_3` is ignored.
    #  If we would explore the state machine, we would still scan `STATE_2`.
    #  However, because `STATE_2` is also reachable from `STATE_3` it will now be
    #  ignored. In most cases this should be fine, but we have to handle it.
    states_to_scan.difference_update(ign_states)
    for ign_state in ign_states:
        states_to_scan.difference_update(reachable_states.get(ign_state, set()))
    assert start_state in states_to_scan

    for downstream_state in states_to_scan:
        if downstream_state in ign_states:
            continue
        for dnode in downstream_state.data_nodes():
            if dnode.data != data_to_look:
                continue
            if dnode in ign_dnodes:
                continue
            if downstream_state.out_degree(dnode) != 0:
                return True  # There is a read operation
    return False


def is_reachable(
    start: Union[dace_nodes.Node, Sequence[dace_nodes.Node]],
    target: Union[dace_nodes.Node, Sequence[dace_nodes.Node]],
    state: dace.SDFGState,
) -> bool:
    """Explores the graph from `start` and checks if `target` is reachable.

    The exploration of the graph is done in a way that ignores the connector names.
    It is possible to pass multiple start nodes and targets. In case of multiple target
    nodes, the function returns True if any of them is reachable.

    Args:
        start: The node from where to start.
        target: The node to look for.
        state: The SDFG state on which we operate.
    """
    to_visit: list[dace_nodes.Node] = [start] if isinstance(start, dace_nodes.Node) else list(start)
    targets: set[dace_nodes.Node] = {target} if isinstance(target, dace_nodes.Node) else set(target)
    seen: set[dace_nodes.Node] = set()

    while to_visit:
        node = to_visit.pop()
        if node in targets:
            return True
        seen.add(node)
        to_visit.extend(oedge.dst for oedge in state.out_edges(node) if oedge.dst not in seen)

    return False


def is_source_node_of(
    sink: dace_nodes.Node,
    possible_sources: Union[dace_nodes.Node, Sequence[dace_nodes.Node]],
    state: dace.SDFGState,
) -> bool:
    """Explores the graph and checks if `possible_sources` produce data for `sink`.

    The function is similar to `is_reachable()` except that it explores the
    graph in the reverse direction of the data flow. The function starts a
    revers depth first search at `sink` and checks if it is possible to reach
    any nodes listed in `possible_sources`.

    Args:
        sink: Start of the reverse DFS.
        possible_sources: Look for these nodes.
        state: The SDFG state on which we operate.
    """
    # TODO(phimuell): Think if this function could be implemented in terms of `is_reachable()`.
    to_visit: list[dace_nodes.Node] = [sink] if isinstance(sink, dace_nodes.Node) else list(sink)
    possible_sources = (
        {possible_sources}
        if isinstance(possible_sources, dace_nodes.Node)
        else set(possible_sources)
    )
    seen: set[dace_nodes.Node] = set()

    while to_visit:
        node = to_visit.pop()
        if node in possible_sources:
            return True
        seen.add(node)
        to_visit.extend(iedge.src for iedge in state.in_edges(node) if iedge.src not in seen)

    return False


def is_view(
    node: Union[dace_nodes.AccessNode, dace_data.Data],
    sdfg: Optional[dace.SDFG] = None,
) -> bool:
    """Tests if `node` points to a view or not."""
    if isinstance(node, dace_nodes.AccessNode):
        assert sdfg is not None
        node_desc = node.desc(sdfg)
    else:
        assert isinstance(node, dace_data.Data)
        node_desc = node
    return isinstance(node_desc, dace_data.View)


def track_view(
    view: dace_nodes.AccessNode,
    state: dace.SDFGState,
    sdfg: dace.SDFG,
) -> dace_nodes.AccessNode:
    """Find the original data of a View.

    Given the View `view`, the function will trace the view back to the original
    access node. For convenience, if `view` is not a `View` the argument will be
    returned.

    Args:
        view: The view that should be traced.
        state: The state in which we operate.
        sdfg: The SDFG on which we operate.
    """

    # Test if it is a view at all, if not return the passed node as source.
    if not is_view(view, sdfg):
        return view

    # First determine if the view is used for reading or writing.
    curr_edge = dace.sdfg.utils.get_view_edge(state, view)
    if curr_edge is None:
        raise RuntimeError(f"Failed to determine the direction of the view '{view}'.")
    if curr_edge.dst_conn == "views":
        # The view is used for reading.
        next_node = lambda curr_edge: curr_edge.src  # noqa: E731 [lambda-assignment]
    elif curr_edge.src_conn == "views":
        # The view is used for writing.
        next_node = lambda curr_edge: curr_edge.dst  # noqa: E731 [lambda-assignment]
    else:
        raise RuntimeError(f"Failed to determine the direction of the view '{view}' | {curr_edge}.")

    # Now trace the view back.
    org_view = view
    view = next_node(curr_edge)
    while is_view(view, sdfg):
        curr_edge = dace.sdfg.utils.get_view_edge(state, view)
        if curr_edge is None:
            raise RuntimeError(f"View tracing of '{org_view}' failed at note '{view}'.")
        view = next_node(curr_edge)
    return view


def reroute_edge(
    is_producer_edge: bool,
    current_edge: dace_graph.MultiConnectorEdge,
    ss_offset: Sequence[dace_sym.SymExpr],
    state: dace.SDFGState,
    sdfg: dace.SDFG,
    old_node: dace_nodes.AccessNode,
    new_node: dace_nodes.AccessNode,
) -> dace_graph.MultiConnectorEdge:
    """Performs the rerouting of the edge.

    Essentially the function create a new edge that replaces `old_node` with `new_node`.
    Depending on the value of `is_producer_edge` the behaviour is slightly different.

    If `is_producer_edge` is `True` then the function expects that `current_edge`
    is an edge that _ends_ at `old_node`. It will then create a new edge that starts
    at the source of `current_edge` but will end in `new_node` and have a similar
    Memlet than `current_edge`.
    If `is_producer_edge` is `False` then it assumes `current_edge` starts at
    `old_node`. It will then create a new edge that starts at `new_node` and has a
    similar Memlet than the original one.
    To modify the subsets where `new_node` is read/written the `ss_offset` is
    provided, this value is added to the current subset.

    It is important that the function will **not** do the following things:
    - Remove the old edge, i.e. `producer_edge`.
    - Modify the data flow at the other side of the edge, to account of the change
        from `old_node` to `new_node`. If this should be modified see
        `reconfigure_dataflow_after_rerouting()`.

    The function returns the new edge.

    Args:
        is_producer_edge: Indicates how to interpret `current_edge`.
        current_edge: The current edge that should be replaced.
        ss_offset: Offset that describes how much to shift writes and reads,
            that were previously associated with `old_node`.
        state: The state in which we operate.
        sdfg: The SDFG on which we operate on.
        old_node: The old that should be replaced in `current_edge`.
        new_node: The new node that should be used instead of `old_node`.
    """
    current_memlet: dace.Memlet = current_edge.data
    if is_producer_edge:
        # NOTE: See the note in `_reconfigure_dataflow()` why it is not save to
        #  use the `get_{dst, src}_subset()` function, although it would be more
        #  appropriate.
        assert current_edge.dst is old_node
        current_subset: dace_sbs.Range = current_memlet.dst_subset
        new_src = current_edge.src
        new_src_conn = current_edge._src_conn
        new_dst = new_node
        new_dst_conn = None
        assert current_edge.dst_conn is None
    else:
        assert current_edge.src is old_node
        current_subset = current_memlet.src_subset
        new_src = new_node
        new_src_conn = None
        new_dst = current_edge.dst
        new_dst_conn = current_edge.dst_conn
        assert current_edge.src_conn is None

    # If the subset we care about, which is always on the `old_node` side, was not
    #  specified we assume that the whole `old_node` has been written.
    # TODO(edopao): Fix lowering that this does not happens, it happens for example
    #  in `tests/next_tests/integration_tests/feature_tests/ffront_tests/
    #  test_execution.py::test_docstring`.
    if current_subset is None:
        current_subset = dace_sbs.Range.from_array(old_node.desc(sdfg))

    # This is the new Memlet, that we will use. We copy it from the original
    #  Memlet and modify it later.
    new_memlet: dace.Memlet = dace.Memlet.from_memlet(current_memlet)

    # Because we operate on the `subset` and `other_subset` properties directly
    #  we do not need to distinguish between the different directions. Also
    #  in both cases the offset is the same.
    if new_memlet.data == old_node.data:
        new_memlet.data = new_node.data
        new_subset = current_subset.offset_new(ss_offset, negative=False)
        new_memlet.subset = new_subset
    else:
        new_subset = current_subset.offset_new(ss_offset, negative=False)
        new_memlet.other_subset = new_subset

    new_edge = state.add_edge(
        new_src,
        new_src_conn,
        new_dst,
        new_dst_conn,
        new_memlet,
    )
    assert (  # Ensure that the edge has the right direction.
        new_subset is new_edge.data.dst_subset
        if is_producer_edge
        else new_subset is new_edge.data.src_subset
    )
    return new_edge


def reconfigure_dataflow_after_rerouting(
    is_producer_edge: bool,
    new_edge: dace_graph.MultiConnectorEdge,
    ss_offset: Sequence[dace_sym.SymExpr] | None,
    state: dace.SDFGState,
    sdfg: dace.SDFG,
    old_node: dace_nodes.AccessNode,
    new_node: dace_nodes.AccessNode,
) -> None:
    """Modify the data flow associated to `new_edge`.

    While the `reroute_edge()` function creates a new edge, it does not modify
    the dataflow at the other side of the connection, this is done by this
    function.

    Depending on the value of `is_producer_edge` the function will either modify
    the source of `new_edge` (`True`) or it will modify the data flow associated
    to the destination of `new_edge` (`False`).
    Furthermore, the specific actions depends on what kind of node is on the other
    side. However, essentially the function will modify it to account for the
    change from `old_node` to `new_node`.

    When using this function the user has to consider the following things:
    - It is the caller's responsibility to ensure that this function is not called
        multiple times on the same producer target.
    - The function will not propagate new strides, this must be done separately.
    - The function will not modify the ranges of Maps.

    Args:
        is_producer_edge: If `True` then the source of `new_edge` is processed,
            if `False` then the destination part of `new_edge` is processed.
        new_edge: The newly created edge, essentially the return value of
            `reroute__edge()`.
        ss_offsets: Offset that describes how much to shift subsets associated
            to `old_node` to account that they are now associated to `new_node`.
            If `None` then `new_node` needs to be a scalar.
        state: The state in which we operate.
        sdfg: The SDFG on which we operate.
        old_node: The old that was involved in the old, rerouted, edge.
        new_node: The new node that should be used instead of `old_node`.
    """
    other_node = new_edge.src if is_producer_edge else new_edge.dst

    if isinstance(other_node, dace_nodes.AccessNode):
        # There is nothing here to do.
        pass

    elif isinstance(other_node, dace_nodes.Tasklet):
        # A very obscure case, but I think it might happen, but as in the AccessNode
        #  case there is nothing to do here.
        pass

    elif isinstance(other_node, (dace_nodes.MapExit | dace_nodes.MapEntry)):
        # Essentially, we have to propagate the change that everything that
        #  refers to `old_node` should now refer to `new_node`, In addition we also
        #  have to modify the subsets, depending on the direction of the new edge
        #  either the source or destination subset.
        # NOTE: Also for this case we have to propagate the strides, for the case
        #   that a NestedSDFG is inside the map, but this is done externally.
        assert (
            isinstance(other_node, dace_nodes.MapExit)
            if is_producer_edge
            else isinstance(other_node, dace_nodes.MapEntry)
        )
        if ss_offset is None:
            if not isinstance(new_node.desc(sdfg), dace_data.Scalar):
                raise TypeError(
                    "Passed 'None' as 'ss_offset' but 'new_node' ({new_node.data}) is not a scalar."
                )

        for memlet_tree in state.memlet_tree(new_edge).traverse_children(include_self=False):
            edge_to_adjust = memlet_tree.edge
            memlet_to_adjust = edge_to_adjust.data

            # If needed modify the association of the Memlet.
            if memlet_to_adjust.data == old_node.data:
                memlet_to_adjust.data = new_node.data

            if ss_offset is None:
                # Scalar modification
                if is_producer_edge:
                    memlet_to_adjust.dst_subset = "0"
                else:
                    memlet_to_adjust.src_subset = "0"

            else:
                # NOTE: Actually we should use the `get_{src, dst}_subset()` functions,
                #  see https://github.com/spcl/dace/issues/1703. However, we can not
                #  do that because the SDFG is currently in an invalid state. So
                #  we have to call the properties and hope that it works.
                subset_to_adjust = (
                    memlet_to_adjust.dst_subset if is_producer_edge else memlet_to_adjust.src_subset
                )
                assert subset_to_adjust is not None
                subset_to_adjust.offset(ss_offset, negative=False)

    elif isinstance(other_node, dace_nodes.NestedSDFG):
        # We have obviously to adjust the strides, however, this is done outside
        #  this function.
        # TODO(phimuell): Look into the implication that we not necessarily pass
        #  the full array, but essentially slice a bit.
        pass

    else:
        # As we encounter them we should handle them case by case.
        raise NotImplementedError(
            f"The case for '{type(other_node).__name__}' has not been implemented."
        )


def find_upstream_nodes(
    start: dace_nodes.Node,
    state: dace.SDFGState,
    start_connector: Optional[str] = None,
    limit_node: Optional[dace_nodes.Node] = None,
) -> set[dace_nodes.Node]:
    """Finds all upstream nodes, i.e. all producers, of `start`.

    Args:
        start: Start the exploring from this node.
        state: The state in which it should be explored.
        start_connector: If given then only consider edges that terminate
            in this connector, otherwise consider all incoming edges.
        limit_node: Consider this node as "limiting wall", i.e. do not explore
            beyond it.
    """
    seen: set[dace_nodes.Node] = set()

    to_visit = [
        iedge.src
        for iedge in (
            state.in_edges(start)
            if start_connector is None
            else state.in_edges_by_connector(start, start_connector)
        )
        if iedge.src is not limit_node
    ]

    while len(to_visit) != 0:
        node = to_visit.pop()
        if node in seen:
            continue
        seen.add(node)
        to_visit.extend(iedge.src for iedge in state.in_edges(node) if iedge.src is not limit_node)

    return seen


def find_successor_state(state: dace.SDFGState) -> list[dace.SDFGState]:
    """Finds the set of next states after `state`.

    In the simplest case the function will return something like this.
    `[e.dst for e in state.sdfg.out_edges(state)`].
    If this set is empty, the function will go up the hierarchy of the control flow
    regions and inspect the outputs. This means if `state` is a terminal node of
    a control flow region the function will find the successor states of the
    region.
    If the set of successors contains an control flow region, the region will be
    expanded, such that the result set only consists of `SDFGState`s.
    """

    def _impl(
        state: Union[dace.SDFGState, dace.sdfg.state.AbstractControlFlowRegion],
        graph: dace.sdfg.state.AbstractControlFlowRegion,
    ) -> list[dace.sdfg.state.AbstractControlFlowRegion]:
        assert state is not graph
        assert state in graph.nodes()

        curr_out_edges = [oedge.dst for oedge in graph.out_edges(state)]

        # End recursion if we found some successor edges or we have reached the top.
        if len(curr_out_edges) > 0 or graph.parent_graph is state.sdfg:
            return curr_out_edges
        elif isinstance(graph.parent_graph, dace.sdfg.state.ConditionalBlock):
            # For conditional we go two levels up, because there is nothing
            #  interesting inside it.
            return _impl(graph.parent_graph, graph.parent_graph.parent_graph)
        else:
            # We have not found any successor edges, so we go one level up.
            #  This time we look for the enclosing graph.
            # NOTE: We do not handle the `LoopRegion` specially, as the successor
            #   state could either be the one following the region or the first
            #   state of the loop body.
            return _impl(graph, graph.parent_graph)

    # Expand the successors such that it contains only `SDFGState`s.
    # TODO(phimuell): Only use the starting block instead of all.
    all_successors: list[dace.SDFGState] = []
    for successor in _impl(state, state.parent_graph):
        if isinstance(successor, dace.SDFGState):
            all_successors.append(successor)
        else:
            all_successors.extend(successor.all_states())

    return all_successors



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import abc
import typing
from typing import Any, Optional

from gt4py.next.iterator import ir as itir
from gt4py.next.type_system import type_specifications as ts


@typing.runtime_checkable
class GTCallable(typing.Protocol):
    """
    Typing Protocol (abstract base class) defining the interface for subroutines.

    Any class implementing the methods defined in this protocol can be called
    from ``ffront`` programs or operators.
    """

    def __gt_closure_vars__(self) -> Optional[dict[str, Any]]:
        """
        Return all external variables referenced inside the callable.

        Note that in addition to the callable itself all captured variables
        are also lowered such that they can be used in the lowered callable.
        """
        return None

    @abc.abstractmethod
    def __gt_type__(self) -> ts.CallableType:
        """
        Return symbol type, i.e. signature and return type.

        The type is used internally to populate the closure vars of the
        various dialects root nodes (i.e. FOAST Field Operator, PAST Program)
        """
        ...

    @abc.abstractmethod
    def __gt_itir__(self) -> itir.FunctionDefinition:
        """
        Return iterator IR function definition representing the callable.

        Used internally by the Program decorator to populate the function
        definitions of the iterator IR.
        """
        ...

    # FIXME[#1582](tehrengruber): remove after refactoring to GTIR
    @abc.abstractmethod
    def __gt_gtir__(self) -> itir.FunctionDefinition:
        """
        Return iterator IR function definition representing the callable.
        Used internally by the Program decorator to populate the function
        definitions of the iterator IR.
        """
        ...

    # TODO(tehrengruber): For embedded execution a `__call__` method and for
    #  "truly" embedded execution arguably also a `from_function` method is
    #  required. Since field operators currently have a `__gt_type__` with a
    #  Field return value, but it's `__call__` method being void (result via
    #  out arg) there is no good / consistent definition on what signature a
    #  protocol implementer is expected to provide. Skipping for now.



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import functools
import itertools
from typing import Any, Callable, ClassVar, Optional, ParamSpec, TypeGuard, TypeVar, cast, overload


class RecursionGuard:
    """
    Context manager to guard against inifinite recursion.

    >>> def foo(i):
    ...     with RecursionGuard(i):
    ...         if i % 2 == 0:
    ...             foo(i)
    ...     return i
    >>> foo(3)
    3
    >>> foo(2)  # doctest:+ELLIPSIS
    Traceback (most recent call last):
        ...
    gt4py.next.utils.RecursionGuard.RecursionDetected
    """

    guarded_objects: ClassVar[set[int]] = set()

    obj: Any

    class RecursionDetected(Exception):
        pass

    def __init__(self, obj: Any):
        self.obj = obj

    def __enter__(self) -> None:
        if id(self.obj) in self.guarded_objects:
            raise self.RecursionDetected()
        self.guarded_objects.add(id(self.obj))

    def __exit__(self, *exc: Any) -> None:
        self.guarded_objects.remove(id(self.obj))


_T = TypeVar("_T")
_P = ParamSpec("_P")
_R = TypeVar("_R")


def is_tuple_of(v: Any, t: type[_T]) -> TypeGuard[tuple[_T, ...]]:
    return isinstance(v, tuple) and all(isinstance(e, t) for e in v)


# TODO(havogt): remove flatten duplications in the whole codebase
def flatten_nested_tuple(
    value: tuple[
        _T | tuple, ...
    ],  # `_T` omitted on purpose as type of `value`, to properly deduce `_T` on the user-side
) -> tuple[_T, ...]:
    if isinstance(value, tuple):
        return sum((flatten_nested_tuple(v) for v in value), start=())  # type: ignore[arg-type] # cannot properly express nesting
    else:
        return (value,)


@overload
def tree_map(
    fun: Callable[_P, _R],
    *,
    collection_type: type | tuple[type, ...] = tuple,
    result_collection_constructor: Optional[type | Callable] = None,
) -> Callable[..., _R | tuple[_R | tuple, ...]]: ...


@overload
def tree_map(
    *,
    collection_type: type | tuple[type, ...] = tuple,
    result_collection_constructor: Optional[type | Callable] = None,
) -> Callable[
    [Callable[_P, _R]], Callable[..., Any]
]: ...  # TODO(havogt): if result_collection_constructor is Callable, improve typing


def tree_map(
    fun: Optional[Callable[_P, _R]] = None,
    *,
    collection_type: type | tuple[type, ...] = tuple,
    result_collection_constructor: Optional[type | Callable] = None,
) -> Callable[..., _R | tuple[_R | tuple, ...]] | Callable[[Callable[_P, _R]], Callable[..., Any]]:
    """
    Apply `fun` to each entry of (possibly nested) collections (by default `tuple`s).

    Args:
        fun: Function to apply to each entry of the collection.
        collection_type: Type of the collection to be traversed. Can be a single type or a tuple of types.
        result_collection_constructor: Type of the collection to be returned. If `None` the same type as `collection_type` is used.

    Examples:
        >>> tree_map(lambda x: x + 1)(((1, 2), 3))
        ((2, 3), 4)

        >>> tree_map(lambda x, y: x + y)(((1, 2), 3), ((4, 5), 6))
        ((5, 7), 9)

        >>> tree_map(collection_type=list)(lambda x: x + 1)([[1, 2], 3])
        [[2, 3], 4]

        >>> tree_map(collection_type=list, result_collection_constructor=tuple)(lambda x: x + 1)(
        ...     [[1, 2], 3]
        ... )
        ((2, 3), 4)

        >>> @tree_map
        ... def impl(x):
        ...     return x + 1
        >>> impl(((1, 2), 3))
        ((2, 3), 4)
    """

    if result_collection_constructor is None:
        if isinstance(collection_type, tuple):
            raise TypeError(
                "tree_map() requires `result_collection_constructor` when `collection_type` is a tuple of types."
            )
        result_collection_constructor = collection_type

    if fun:

        @functools.wraps(fun)
        def impl(*args: Any | tuple[Any | tuple, ...]) -> _R | tuple[_R | tuple, ...]:
            if isinstance(args[0], collection_type):
                assert all(
                    isinstance(arg, collection_type) and len(args[0]) == len(arg) for arg in args
                )
                assert result_collection_constructor is not None
                return result_collection_constructor(impl(*arg) for arg in zip(*args))

            return fun(  # type: ignore[call-arg, misc] # mypy not smart enough
                *cast(_P.args, args)
            )  # mypy doesn't understand that `args` at this point is of type `_P.args`

        return impl
    else:
        return functools.partial(
            tree_map,
            collection_type=collection_type,
            result_collection_constructor=result_collection_constructor,
        )


_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")


def equalize_tuple_structure(
    d1: _T1, d2: _T2, *, fill_value: Any, bidirectional: bool = True
) -> tuple[_T1, _T2]:
    """
    Given two values or tuples thereof equalize their structure.

    If one of the arguments is a tuple the argument will be promoted to a tuple of the same
    structure.

    >>> equalize_tuple_structure((1,), (2, 3), fill_value=None) == (
    ...     (1, None),
    ...     (2, 3),
    ... )
    True

    >>> equalize_tuple_structure(1, (2, 3), fill_value=None) == (
    ...     (1, 1),
    ...     (2, 3),
    ... )
    True

    If `bidirectional` is `False` only the second argument is equalized and an error is raised if
    the first argument would require promotion.
    """
    d1_is_tuple = isinstance(d1, tuple)
    d2_is_tuple = isinstance(d2, tuple)
    if not d1_is_tuple and d2_is_tuple:
        if not bidirectional:
            raise ValueError(f"Expected `{d2!s}` to have same structure as `{d1!s}`.")
        return equalize_tuple_structure(
            (d1,) * len(d2),  # type: ignore[arg-type] # ensured by condition above
            d2,
            fill_value=fill_value,
            bidirectional=bidirectional,
        )
    if d1_is_tuple and not d2_is_tuple:
        return equalize_tuple_structure(
            d1,
            (d2,) * len(d1),  # type: ignore[arg-type] # ensured by condition above
            fill_value=fill_value,
            bidirectional=bidirectional,
        )
    if d1_is_tuple and d2_is_tuple:
        if not bidirectional and len(d1) < len(d2):  # type: ignore[arg-type] # ensured by condition above
            raise ValueError(f"Expected `{d2!s}` to be longer than or equal to `{d1!s}`.")
        return tuple(  # type: ignore[return-value] # mypy not smart enough
            zip(
                *(
                    equalize_tuple_structure(
                        el1, el2, fill_value=fill_value, bidirectional=bidirectional
                    )
                    for el1, el2 in itertools.zip_longest(d1, d2, fillvalue=fill_value)  # type: ignore[call-overload] # d1, d2 are tuples
                )
            )
        )
    return d1, d2



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import enum
import functools
import operator
from typing import Optional

from gt4py import eve
from gt4py._core import definitions as core_defs
from gt4py.next.iterator import builtins, embedded, ir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, ir_makers as im
from gt4py.next.iterator.transforms import fixed_point_transformation
from gt4py.next.type_system import type_specifications as ts


def _value_from_literal(literal: ir.Literal) -> core_defs.Scalar:
    if literal.type.kind == ts.ScalarKind.BOOL:
        assert literal.value in ["True", "False"]
        return literal.value == "True"
    return getattr(embedded, str(literal.type))(literal.value)


class UndoCanonicalizeMinus(eve.NodeTranslator):
    PRESERVED_ANNEX_ATTRS = (
        "type",
        "domain",
    )

    def visit_FunCall(self, node: ir.FunCall, **kwargs) -> ir.Node:
        node = super().generic_visit(node, **kwargs)
        # `a + (-b)` -> `a - b` , `-a + b` -> `b - a`, `-a + (-b)` -> `-a - b`
        if cpm.is_call_to(node, "plus"):
            a, b = node.args
            if cpm.is_call_to(b, "neg"):
                return im.minus(a, b.args[0])
            if isinstance(b, ir.Literal) and (val := _value_from_literal(b)) < 0:
                return im.minus(a, -val)  # type: ignore[operator] # if val would represent an unsigend int, `-` is not supported, but error would be somewhere else
            if cpm.is_call_to(a, "neg"):
                return im.minus(b, a.args[0])
            if isinstance(a, ir.Literal) and (val := _value_from_literal(a)) < 0:
                return im.minus(b, -val)  # type: ignore[operator] # if val would represent an unsigend int, `-` is not supported, but error would be somewhere else
        return node


_COMMUTATIVE_OPS = ("plus", "multiplies", "minimum", "maximum")


@dataclasses.dataclass(frozen=True, kw_only=True)
class ConstantFolding(
    fixed_point_transformation.CombinedFixedPointTransform, eve.PreserveLocationVisitor
):
    PRESERVED_ANNEX_ATTRS = (
        "type",
        "domain",
    )

    class Transformation(enum.Flag):
        # `1 + a` -> `a + 1`, prerequisite for FOLD_FUNCALL_LITERAL, FOLD_NEUTRAL_OP
        # `1 + f(...)` -> `f(...) + 1`, prerequisite for FOLD_FUNCALL_LITERAL, FOLD_NEUTRAL_OP
        # `f(...) + (expr1 + expr2)` -> `(expr1 + expr2) + f(...)`, for `s[0] + (s[0] + 1)`, prerequisite for FOLD_MIN_MAX_PLUS
        CANONICALIZE_OP_FUNCALL_SYMREF_LITERAL = enum.auto()

        # `a - b` -> `a + (-b)`, prerequisite for FOLD_MIN_MAX_PLUS
        CANONICALIZE_MINUS = enum.auto()

        # `maximum(a, maximum(...))` -> `maximum(maximum(...), a)`, prerequisite for FOLD_MIN_MAX
        CANONICALIZE_MIN_MAX = enum.auto()

        # `(a + 1) + 1` -> `a + (1 + 1)`
        FOLD_FUNCALL_LITERAL = enum.auto()

        # `maximum(maximum(a, 1), a)` -> `maximum(a, 1)`
        # `maximum(maximum(a, 1), 1)` -> `maximum(a, 1)`
        FOLD_MIN_MAX = enum.auto()

        # `maximum(a + 1), a)` -> `a + 1`
        # `maximum(a + 1, a + (-1))` -> `a + maximum(1, -1)`
        FOLD_MIN_MAX_PLUS = enum.auto()

        # `a + 0` -> `a`, `a * 1` -> `a`
        FOLD_NEUTRAL_OP = enum.auto()

        # `1 + 1` -> `2`
        FOLD_ARITHMETIC_BUILTINS = enum.auto()

        # `minimum(a, a)` -> `a`
        FOLD_MIN_MAX_LITERALS = enum.auto()

        # `if_(True, true_branch, false_branch)` -> `true_branch`
        FOLD_IF = enum.auto()

        FOLD_INFINITY_ARITHMETIC = enum.auto()

        @classmethod
        def all(self) -> ConstantFolding.Transformation:
            return functools.reduce(operator.or_, self.__members__.values())

    enabled_transformations: Transformation = Transformation.all()  # noqa: RUF009 [function-call-in-dataclass-default-argument]

    @classmethod
    def apply(
        cls, node: ir.Node, enabled_transformations: Optional[Transformation] = None
    ) -> ir.Node:
        enabled_transformations = enabled_transformations or cls.Transformation.all()

        node = cls(enabled_transformations=enabled_transformations).visit(node)
        return UndoCanonicalizeMinus().visit(node)

    def transform_canonicalize_op_funcall_symref_literal(
        self, node: ir.FunCall, **kwargs
    ) -> Optional[ir.Node]:
        # `op(literal, symref|funcall)` -> `op(symref|funcall, literal)`
        # `op1(funcall, op2(...))` -> `op1(op2(...), funcall)` for `s[0] + (s[0] + 1)`
        if cpm.is_call_to(node, _COMMUTATIVE_OPS):
            a, b = node.args
            if (isinstance(a, ir.Literal) and not isinstance(b, ir.Literal)) or (
                not cpm.is_call_to(a, _COMMUTATIVE_OPS) and cpm.is_call_to(b, _COMMUTATIVE_OPS)
            ):
                return im.call(node.fun)(b, a)
        return None

    def transform_canonicalize_minus(self, node: ir.FunCall, **kwargs) -> Optional[ir.Node]:
        # `a - b` -> `a + (-b)`
        if cpm.is_call_to(node, "minus"):
            return im.plus(node.args[0], self.fp_transform(im.call("neg")(node.args[1])))
        return None

    def transform_canonicalize_min_max(self, node: ir.FunCall, **kwargs) -> Optional[ir.Node]:
        # `maximum(a, maximum(...))` -> `maximum(maximum(...), a)`
        if cpm.is_call_to(node, ("maximum", "minimum")):
            op = node.fun.id
            if cpm.is_call_to(node.args[1], op) and not cpm.is_call_to(node.args[0], op):
                return im.call(op)(node.args[1], node.args[0])
        return None

    def transform_fold_funcall_literal(self, node: ir.FunCall, **kwargs) -> Optional[ir.Node]:
        # `(a + 1) + 1` -> `a + (1 + 1)`
        if cpm.is_call_to(node, "plus"):
            if cpm.is_call_to(node.args[0], "plus") and isinstance(node.args[1], ir.Literal):
                (expr, lit1), lit2 = node.args[0].args, node.args[1]
                if isinstance(expr, (ir.SymRef, ir.FunCall)) and isinstance(lit1, ir.Literal):
                    return im.plus(
                        expr,
                        self.fp_transform(im.plus(lit1, lit2)),
                    )
        return None

    def transform_fold_min_max(self, node: ir.FunCall, **kwargs) -> Optional[ir.Node]:
        # `maximum(maximum(a, 1), a)` -> `maximum(a, 1)`
        # `maximum(maximum(a, 1), 1)` -> `maximum(a, 1)`
        if cpm.is_call_to(node, ("minimum", "maximum")):
            op = node.fun.id
            if cpm.is_call_to(node.args[0], op):
                fun_call, arg1 = node.args
                if arg1 in fun_call.args:  # type: ignore[attr-defined] # assured by if above
                    return fun_call
        return None

    def transform_fold_min_max_plus(self, node: ir.FunCall, **kwargs) -> Optional[ir.Node]:
        if (
            isinstance(node, ir.FunCall)
            and isinstance(node.fun, ir.SymRef)
            and cpm.is_call_to(node, ("minimum", "maximum"))
        ):
            arg0, arg1 = node.args
            # `maximum(a + 1, a)` -> `a + 1`
            if cpm.is_call_to(arg0, "plus"):
                if arg0.args[0] == arg1:
                    return im.plus(
                        arg0.args[0], self.fp_transform(im.call(node.fun.id)(arg0.args[1], 0))
                    )
            # `maximum(a + 1, a + (-1))` -> `a + maximum(1, -1)`
            if cpm.is_call_to(arg0, "plus") and cpm.is_call_to(arg1, "plus"):
                if arg0.args[0] == arg1.args[0]:
                    return im.plus(
                        arg0.args[0],
                        self.fp_transform(im.call(node.fun.id)(arg0.args[1], arg1.args[1])),
                    )

        return None

    def transform_fold_neutral_op(self, node: ir.FunCall, **kwargs) -> Optional[ir.Node]:
        # `a + 0` -> `a`, `a * 1` -> `a`
        if (
            cpm.is_call_to(node, "plus")
            and isinstance(node.args[1], ir.Literal)
            and node.args[1].value.isdigit()
            and int(node.args[1].value) == 0
        ) or (
            cpm.is_call_to(node, "multiplies")
            and isinstance(node.args[1], ir.Literal)
            and node.args[1].value.isdigit()
            and int(node.args[1].value) == 1
        ):
            return node.args[0]
        return None

    @classmethod
    def transform_fold_arithmetic_builtins(self, node: ir.FunCall, **kwargs) -> Optional[ir.Node]:
        # `1 + 1` -> `2`
        if (
            isinstance(node, ir.FunCall)
            and isinstance(node.fun, ir.SymRef)
            and len(node.args) > 0
            and all(isinstance(arg, ir.Literal) for arg in node.args)
        ):
            try:
                if node.fun.id in builtins.ARITHMETIC_BUILTINS:
                    fun = getattr(embedded, str(node.fun.id))
                    arg_values = [
                        _value_from_literal(arg)  # type: ignore[arg-type] # arg type already established in if condition
                        for arg in node.args
                    ]
                    return im.literal_from_value(fun(*arg_values))
            except ValueError:
                pass  # happens for inf and neginf
        return None

    def transform_fold_min_max_literals(self, node: ir.FunCall, **kwargs) -> Optional[ir.Node]:
        # `minimum(a, a)` -> `a`
        if cpm.is_call_to(node, ("minimum", "maximum")):
            if node.args[0] == node.args[1]:
                return node.args[0]
        return None

    def transform_fold_if(self, node: ir.FunCall, **kwargs) -> Optional[ir.Node]:
        # `if_(True, true_branch, false_branch)` -> `true_branch`
        if cpm.is_call_to(node, "if_") and isinstance(node.args[0], ir.Literal):
            if node.args[0].value == "True":
                return node.args[1]
            else:
                assert node.args[0].value == "False"
                return node.args[2]
        return None

    def transform_fold_infinity_arithmetic(self, node: ir.FunCall) -> Optional[ir.Node]:
        if cpm.is_call_to(node, "plus"):
            # `a + +/-inf` -> `+/-inf`
            a, b = node.args
            assert not (isinstance(a, ir.InfinityLiteral) and isinstance(b, ir.InfinityLiteral))
            for arg in a, b:
                if isinstance(arg, ir.InfinityLiteral):
                    return arg

        if cpm.is_call_to(node, "minimum"):
            if ir.InfinityLiteral.NEGATIVE in node.args:
                # `minimum(-inf, a)` -> `-inf`
                return ir.InfinityLiteral.NEGATIVE
            if ir.InfinityLiteral.POSITIVE in node.args:
                # `minimum(inf, a)` -> `a`
                a, b = node.args
                return b if a == ir.InfinityLiteral.POSITIVE else a

        if cpm.is_call_to(node, "maximum"):
            if ir.InfinityLiteral.POSITIVE in node.args:
                # `maximum(inf, a)` -> `inf`
                return ir.InfinityLiteral.POSITIVE
            if ir.InfinityLiteral.NEGATIVE in node.args:
                # `maximum(-inf, a)` -> `a`
                a, b = node.args
                return b if a == ir.InfinityLiteral.NEGATIVE else a

        if cpm.is_call_to(node, ("less", "less_equal")):
            a, b = node.args
            # we don't handle `inf < inf` or `-inf < -inf`.args
            assert a != b or not isinstance(a, ir.InfinityLiteral)

            # `-inf < v` -> `True`
            # `v < inf` -> `True`
            if a == ir.InfinityLiteral.NEGATIVE or b == ir.InfinityLiteral.POSITIVE:
                return im.literal_from_value(True)
            # `inf < v` -> `False`
            # `v < -inf ` -> `False`
            if a == ir.InfinityLiteral.POSITIVE or b == ir.InfinityLiteral.NEGATIVE:
                return im.literal_from_value(False)

        if cpm.is_call_to(node, ("greater", "greater_equal")):
            a, b = node.args
            # we don't handle `inf > inf` or `-inf > -inf`.args
            assert a != b or not isinstance(a, ir.InfinityLiteral)

            # `inf > v` -> `True`
            # `v > -inf ` -> `True`
            if a == ir.InfinityLiteral.POSITIVE or b == ir.InfinityLiteral.NEGATIVE:
                return im.literal_from_value(True)
            # `-inf > v` -> `False`
            # `v > inf` -> `False`
            if a == ir.InfinityLiteral.NEGATIVE or b == ir.InfinityLiteral.POSITIVE:
                return im.literal_from_value(False)

        return None



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from gt4py.next import backend as next_backend
from gt4py.next.program_processors.runners import roundtrip


backend = next_backend.Backend(
    name="double_roundtrip",
    transforms=next_backend.DEFAULT_TRANSFORMS,
    executor=roundtrip.Roundtrip(dispatch_backend=roundtrip.default, use_embedded=False),
    allocator=roundtrip.default.allocator,
)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import abc
import dataclasses
import functools

import gt4py._core.definitions as core_defs
import gt4py.next.common as common
import gt4py.storage.allocators as core_allocators
from gt4py.eve.extended_typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Final,
    Optional,
    Protocol,
    Sequence,
    TypeAlias,
    TypeGuard,
    cast,
)


FieldLayoutMapper: TypeAlias = Callable[
    [Sequence[common.Dimension]], core_allocators.BufferLayoutMap
]


class FieldBufferAllocatorProtocol(Protocol[core_defs.DeviceTypeT]):
    """Protocol for buffer allocators used to allocate memory for fields with a given domain."""

    @property
    @abc.abstractmethod
    def __gt_device_type__(self) -> core_defs.DeviceTypeT: ...

    @abc.abstractmethod
    def __gt_allocate__(
        self,
        domain: common.Domain,
        dtype: core_defs.DType[core_defs.ScalarT],
        device_id: int = 0,
        aligned_index: Optional[Sequence[common.NamedIndex]] = None,  # absolute position
    ) -> core_allocators.TensorBuffer[core_defs.DeviceTypeT, core_defs.ScalarT]: ...


def is_field_allocator(obj: Any) -> TypeGuard[FieldBufferAllocatorProtocol]:
    return hasattr(obj, "__gt_device_type__") and hasattr(obj, "__gt_allocate__")


def is_field_allocator_for(
    obj: Any, device: core_defs.DeviceTypeT
) -> TypeGuard[FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]]:
    return is_field_allocator(obj) and obj.__gt_device_type__ is device


class FieldBufferAllocatorFactoryProtocol(Protocol[core_defs.DeviceTypeT]):
    """Protocol for device-specific buffer allocator factories for fields."""

    @property
    @abc.abstractmethod
    def __gt_allocator__(self) -> FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]: ...


def is_field_allocator_factory(obj: Any) -> TypeGuard[FieldBufferAllocatorFactoryProtocol]:
    return hasattr(obj, "__gt_allocator__")


def is_field_allocator_factory_for(
    obj: Any, device: core_defs.DeviceTypeT
) -> TypeGuard[FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]]:
    return is_field_allocator_factory(obj) and obj.__gt_allocator__.__gt_device_type__ is device


FieldBufferAllocationUtil = (
    FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]
    | FieldBufferAllocatorFactoryProtocol[core_defs.DeviceTypeT]
)


def is_field_allocation_tool(obj: Any) -> TypeGuard[FieldBufferAllocationUtil]:
    return is_field_allocator(obj) or is_field_allocator_factory(obj)


def get_allocator(
    obj: Any,
    *,
    default: Optional[FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]] = None,
    strict: bool = False,
) -> Optional[FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]]:
    """
    Return a field-buffer-allocator from an object assumed to be an allocator or an allocator factory.

    A default allocator can be provided as fallback in case `obj` is neither an allocator nor a factory.

    Arguments:
        obj: The allocator or allocator factory.
        default: Fallback allocator.
        strict: If `True`, raise an exception if there is no way to get a valid allocator
            from `obj` or `default`.

    Returns:
        A field buffer allocator.

    Raises:
        TypeError: If `obj` is neither a field allocator nor a field allocator factory and no default
            is provided in `strict` mode.
    """
    if is_field_allocator(obj):
        return obj
    elif is_field_allocator_factory(obj):
        return obj.__gt_allocator__
    elif not strict or is_field_allocator(default):
        return default
    else:
        raise TypeError(
            f"Object '{obj}' is neither a field allocator nor a field allocator factory."
        )


@dataclasses.dataclass(frozen=True)
class BaseFieldBufferAllocator(FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]):
    """Parametrizable field buffer allocator base class."""

    device_type: core_defs.DeviceTypeT
    array_utils: core_allocators.ArrayUtils
    layout_mapper: FieldLayoutMapper
    byte_alignment: int

    @functools.cached_property
    def buffer_allocator(self) -> core_allocators.BufferAllocator:
        return core_allocators.NDArrayBufferAllocator(self.device_type, self.array_utils)

    @property
    def __gt_device_type__(self) -> core_defs.DeviceTypeT:
        return self.device_type

    def __gt_allocate__(
        self,
        domain: common.Domain,
        dtype: core_defs.DType[core_defs.ScalarT],
        device_id: int = 0,
        aligned_index: Optional[Sequence[common.NamedIndex]] = None,  # absolute position
    ) -> core_allocators.TensorBuffer[core_defs.DeviceTypeT, core_defs.ScalarT]:
        shape = domain.shape
        layout_map = self.layout_mapper(domain.dims)
        # TODO(egparedes): add support for non-empty aligned index values
        assert aligned_index is None

        return self.buffer_allocator.allocate(
            shape, dtype, device_id, layout_map, self.byte_alignment, aligned_index
        )


if TYPE_CHECKING:
    __TensorFieldAllocatorAsFieldAllocatorInterfaceT: type[FieldBufferAllocatorProtocol] = (
        BaseFieldBufferAllocator
    )


def horizontal_first_layout_mapper(
    dims: Sequence[common.Dimension],
) -> core_allocators.BufferLayoutMap:
    """Map dimensions to a buffer layout making horizontal dims change the slowest (i.e. larger strides)."""

    def pos_of_kind(kind: common.DimensionKind) -> list[int]:
        return [i for i, dim in enumerate(dims) if dim.kind == kind]

    horizontals = pos_of_kind(common.DimensionKind.HORIZONTAL)
    verticals = pos_of_kind(common.DimensionKind.VERTICAL)
    locals_ = pos_of_kind(common.DimensionKind.LOCAL)

    layout_map = [0] * len(dims)
    for i, pos in enumerate(horizontals + verticals + locals_):
        layout_map[pos] = len(dims) - 1 - i

    valid_layout_map = tuple(layout_map)
    assert core_allocators.is_valid_layout_map(valid_layout_map)

    return valid_layout_map


if TYPE_CHECKING:
    __horizontal_first_layout_mapper: FieldLayoutMapper = horizontal_first_layout_mapper


#: Registry of default allocators for each device type.
device_allocators: dict[core_defs.DeviceType, FieldBufferAllocatorProtocol] = {}


class StandardCPUFieldBufferAllocator(BaseFieldBufferAllocator[core_defs.CPUDeviceTyping]):
    """A field buffer allocator for CPU devices that uses a horizontal-first layout mapper and 64-byte alignment."""

    def __init__(self) -> None:
        super().__init__(
            device_type=core_defs.DeviceType.CPU,
            array_utils=core_allocators.numpy_array_utils,
            layout_mapper=horizontal_first_layout_mapper,
            byte_alignment=64,
        )


device_allocators[core_defs.DeviceType.CPU] = StandardCPUFieldBufferAllocator()


assert is_field_allocator(device_allocators[core_defs.DeviceType.CPU])


@dataclasses.dataclass(frozen=True)
class InvalidFieldBufferAllocator(FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]):
    """A field buffer allocator that always raises an exception."""

    device_type: core_defs.DeviceTypeT
    exception: Exception

    @property
    def __gt_device_type__(self) -> core_defs.DeviceTypeT:
        return self.device_type

    def __gt_allocate__(
        self,
        domain: common.Domain,
        dtype: core_defs.DType[core_defs.ScalarT],
        device_id: int = 0,
        aligned_index: Optional[Sequence[common.NamedIndex]] = None,  # absolute position
    ) -> core_allocators.TensorBuffer[core_defs.DeviceTypeT, core_defs.ScalarT]:
        raise self.exception


if core_defs.CUPY_DEVICE_TYPE is not None:
    assert isinstance(core_allocators.cupy_array_utils, core_allocators.ArrayUtils)
    cupy_array_utils = core_allocators.cupy_array_utils

    if core_defs.CUPY_DEVICE_TYPE is core_defs.DeviceType.CUDA:

        class CUDAFieldBufferAllocator(BaseFieldBufferAllocator[core_defs.CUDADeviceTyping]):
            def __init__(self) -> None:
                super().__init__(
                    device_type=core_defs.DeviceType.CUDA,
                    array_utils=cupy_array_utils,
                    layout_mapper=horizontal_first_layout_mapper,
                    byte_alignment=128,
                )

        device_allocators[core_defs.DeviceType.CUDA] = CUDAFieldBufferAllocator()

    else:

        class ROCMFieldBufferAllocator(BaseFieldBufferAllocator[core_defs.ROCMDeviceTyping]):
            def __init__(self) -> None:
                super().__init__(
                    device_type=core_defs.DeviceType.ROCM,
                    array_utils=cupy_array_utils,
                    layout_mapper=horizontal_first_layout_mapper,
                    byte_alignment=128,
                )

        device_allocators[core_defs.DeviceType.ROCM] = ROCMFieldBufferAllocator()

else:

    class InvalidGPUFieldBufferAllocator(InvalidFieldBufferAllocator[core_defs.CUDADeviceTyping]):
        def __init__(self) -> None:
            super().__init__(
                device_type=core_defs.DeviceType.CUDA,
                exception=RuntimeError("Missing `cupy` dependency for GPU allocation"),
            )


StandardGPUFieldBufferAllocator: Final[type[FieldBufferAllocatorProtocol]] = cast(
    type[FieldBufferAllocatorProtocol],
    type(device_allocators[core_defs.CUPY_DEVICE_TYPE])
    if core_defs.CUPY_DEVICE_TYPE
    else InvalidGPUFieldBufferAllocator,
)


def allocate(
    domain: common.DomainLike,
    dtype: core_defs.DType[core_defs.ScalarT],
    *,
    aligned_index: Optional[Sequence[common.NamedIndex]] = None,
    allocator: Optional[FieldBufferAllocationUtil] = None,
    device: Optional[core_defs.Device] = None,
) -> core_allocators.TensorBuffer:
    """
    Allocate a TensorBuffer for the given domain and device or allocator.

    The arguments `device` and `allocator` are mutually exclusive.
    If `device` is specified, the corresponding default allocator
    (defined in :data:`device_allocators`) is used.

    Arguments:
        domain: The domain which should be backed by the allocated tensor buffer.
        dtype: Data type.
        aligned_index: N-dimensional index of the first aligned element
        allocator: The allocator to use for the allocation.
        device: The device to allocate the tensor buffer on (using the default
            allocator for this kind of device from :data:`device_allocators`).

    Returns:
        The allocated tensor buffer.

    Raises:
        ValueError
            If illegal or inconsistent arguments are specified.

    """
    if device is None and allocator is None:
        raise ValueError("No 'device' or 'allocator' specified.")
    actual_allocator = get_allocator(allocator)
    if actual_allocator is None:
        assert device is not None  # for mypy
        actual_allocator = device_allocators[device.device_type]
    elif device is None:
        device = core_defs.Device(actual_allocator.__gt_device_type__, 0)
    elif device.device_type != actual_allocator.__gt_device_type__:
        raise ValueError(f"Device '{device}' and allocator '{actual_allocator}' are incompatible.")

    return actual_allocator.__gt_allocate__(
        domain=common.domain(domain),
        dtype=dtype,
        device_id=device.device_id,
        aligned_index=aligned_index,
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Sequence

import gt4py.eve as eve
from gt4py.eve.codegen import JinjaTemplate as as_jinja
from gt4py.next.otf.binding import interface
from gt4py.next.otf.compilation import common


class FindDependency(eve.Node):
    name: str
    version: str


class LinkDependency(eve.Node):
    name: str
    target: str


class Language(eve.Node):
    name: str


class CMakeListsFile(eve.Node):
    project_name: str
    find_deps: Sequence[FindDependency]
    link_deps: Sequence[LinkDependency]
    source_names: Sequence[str]
    bin_output_suffix: str
    languages: Sequence[Language]


class CMakeListsGenerator(eve.codegen.TemplatedGenerator):
    CMakeListsFile = as_jinja(
        """
        cmake_minimum_required(VERSION 3.20.0)

        project({{project_name}})

        # Languages
        {{"\\n".join(languages)}}

        # Paths
        list(APPEND CMAKE_MODULE_PATH ${CMAKE_BINARY_DIR})
        list(APPEND CMAKE_PREFIX_PATH ${CMAKE_BINARY_DIR})

        set(CMAKE_ARCHIVE_OUTPUT_DIRECTORY ${CMAKE_BINARY_DIR}/bin)
        set(CMAKE_LIBRARY_OUTPUT_DIRECTORY ${CMAKE_BINARY_DIR}/bin)
        set(CMAKE_RUNTIME_OUTPUT_DIRECTORY ${CMAKE_BINARY_DIR}/bin)

        # Find dependencies
        include(FetchContent)

        {{"\\n".join(find_deps)}}

        # Targets
        add_library({{project_name}} MODULE)

        target_compile_features({{project_name}} PRIVATE cxx_std_17)
        set_target_properties({{project_name}} PROPERTIES PREFIX "" SUFFIX ".{{bin_output_suffix}}")

        target_sources({{project_name}}
            PRIVATE
                {{"\\n".join(source_names)}}
        )

        # Link dependencies
        {{"\\n".join(link_deps)}}
        """
    )

    def visit_FindDependency(self, dep: FindDependency) -> str:
        # TODO(ricoh): do not add more libraries here
        #   and do not use this design in a new build system.
        #   Instead, design this to be extensible (refer to ADR-0016).
        match dep.name:
            case "nanobind":
                import sys

                import nanobind

                py = f"""
                set(Python_EXECUTABLE {sys.executable})

                find_package(Python COMPONENTS Interpreter Development REQUIRED)
                """
                nb = f"find_package(nanobind CONFIG REQUIRED PATHS {nanobind.cmake_dir()} NO_DEFAULT_PATHS)"
                return py + "\n" + nb
            case "gridtools_cpu" | "gridtools_gpu":
                import gridtools_cpp

                return f"find_package(GridTools REQUIRED PATHS {gridtools_cpp.get_cmake_dir()} NO_DEFAULT_PATH)"
            case _:
                raise ValueError(f"Library '{dep.name}' is not supported")

    def visit_LinkDependency(self, dep: LinkDependency) -> str:
        # TODO(ricoh): do not add more libraries here
        #   and do not use this design in a new build system.
        #   Instead, design this to be extensible (refer to ADR-0016).
        match dep.name:
            case "nanobind":
                lib_name = "nanobind-static"
            case "gridtools_cpu":
                lib_name = "GridTools::fn_naive"
            case "gridtools_gpu":
                lib_name = "GridTools::fn_gpu"
            case _:
                raise ValueError(f"Library '{dep.name}' is not supported")

        cfg = ""
        if dep.name == "nanobind":
            cfg = "\n".join(
                [
                    "nanobind_build_library(nanobind-static)",
                    f"nanobind_compile_options({dep.target})",
                    f"nanobind_link_options({dep.target})",
                ]
            )
        lnk = f"target_link_libraries({dep.target} PUBLIC {lib_name})"
        return cfg + "\n" + lnk

    Language = as_jinja("enable_language({{name}})")


def generate_cmakelists_source(
    project_name: str,
    dependencies: tuple[interface.LibraryDependency, ...],
    source_names: Sequence[str],
    languages: Sequence[Language] = (Language(name="CXX"),),
) -> str:
    """
    Generate CMakeLists file contents.

    Assumes the name of the gt4py program to be the same as the project name.
    """
    cmakelists_file = CMakeListsFile(
        project_name=project_name,
        find_deps=[FindDependency(name=d.name, version=d.version) for d in dependencies],
        link_deps=[LinkDependency(name=d.name, target=project_name) for d in dependencies],
        source_names=source_names,
        bin_output_suffix=common.python_module_suffix(),
        languages=languages,
    )
    return CMakeListsGenerator.apply(cmakelists_file)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import math
import typing
from typing import Any, Dict, Tuple, Union

from gt4py import eve
from gt4py.cartesian.gtc import gtir
from gt4py.cartesian.gtc.common import LevelMarker


def _iter_field_names(
    node: Union[gtir.Stencil, gtir.ParAssignStmt],
) -> eve.utils.XIterable[gtir.FieldAccess]:
    return node.walk_values().if_isinstance(gtir.FieldDecl).getattr("name").unique()


class KBoundaryVisitor(eve.NodeVisitor):
    """For every field compute the boundary in k, e.g. (2, -1) if [k_origin-2, k_origin+k_domain-1] is accessed."""

    def visit_Stencil(self, node: gtir.Stencil, **kwargs: Any) -> Dict[str, Tuple[int, int]]:
        field_boundaries = {name: (-math.inf, -math.inf) for name in _iter_field_names(node)}
        for vloop in node.vertical_loops:
            self.generic_visit(vloop.body, vloop=vloop, field_boundaries=field_boundaries, **kwargs)
        # if there is no left or right boundary set to zero
        for name, b in field_boundaries.items():
            field_boundaries[name] = (
                b[0] if b[0] != -math.inf else 0,
                b[1] if b[1] != -math.inf else 0,
            )
        return typing.cast(Dict[str, Tuple[int, int]], field_boundaries)

    def visit_FieldAccess(
        self,
        node: gtir.FieldAccess,
        vloop: gtir.VerticalLoop,
        field_boundaries: Dict[str, Tuple[Union[float, int], Union[float, int]]],
        **_: Any,
    ):
        boundary = field_boundaries[node.name]
        interval = vloop.interval
        if not isinstance(node.offset, gtir.VariableKOffset):
            if interval.start.level == LevelMarker.START:
                boundary = (
                    max(-interval.start.offset - node.offset.k, boundary[0]),
                    boundary[1],
                )
            if interval.end.level == LevelMarker.END:
                boundary = (
                    boundary[0],
                    max(interval.end.offset + node.offset.k, boundary[1]),
                )
        if node.name in [decl.name for decl in vloop.temporaries] and (
            boundary[0] > 0 or boundary[1] > 0
        ):
            raise TypeError(f"Invalid access with offset in k to temporary field {node.name}.")
        assert node.name in field_boundaries
        field_boundaries[node.name] = boundary


def compute_k_boundary(node: gtir.Stencil) -> Dict[str, Tuple[int, int]]:
    # loop from START to END is not considered as it might be empty. additional check possible in the future
    return KBoundaryVisitor().visit(node)


def compute_min_k_size(node: gtir.Stencil) -> int:
    """Compute the required number of k levels to run a stencil."""

    min_size_start = 0
    min_size_end = 0
    biggest_offset = 0
    for vloop in node.vertical_loops:
        if (
            vloop.interval.start.level == LevelMarker.START
            and vloop.interval.end.level == LevelMarker.END
        ):
            if not (vloop.interval.start.offset == 0 and vloop.interval.end.offset == 0):
                biggest_offset = max(
                    biggest_offset,
                    vloop.interval.start.offset - vloop.interval.end.offset + 1,
                )
        elif (
            vloop.interval.start.level == LevelMarker.START
            and vloop.interval.end.level == LevelMarker.START
        ):
            min_size_start = max(min_size_start, vloop.interval.end.offset)
            biggest_offset = max(biggest_offset, vloop.interval.end.offset)
        else:
            min_size_end = max(min_size_end, -vloop.interval.start.offset)
            biggest_offset = max(biggest_offset, -vloop.interval.start.offset)
    minimal_size = max(min_size_start + min_size_end, biggest_offset)
    return minimal_size



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Any

import gt4py.next.ffront.field_operator_ast as foast
from gt4py.eve import NodeTranslator, traits
from gt4py.next.type_system import type_translation


@dataclass(frozen=True)
class ClosureVarTypeDeduction(NodeTranslator, traits.VisitorWithSymbolTableTrait):
    """
    Deduce the type of all closure variable declarations.

    After this pass all closure var declarations, i.e. the `Symbol`s contained in
    `foast.FunctionDefinition.closure_vars`, have the type as specified by their
    actual values (given by `closure_vars` to `apply`).
    """

    closure_vars: dict[str, Any]

    @classmethod
    def apply(
        cls, node: foast.FunctionDefinition, closure_vars: dict[str, Any]
    ) -> foast.FunctionDefinition:
        return cls(closure_vars=closure_vars).visit(node)

    def visit_FunctionDefinition(
        self, node: foast.FunctionDefinition, **kwargs: Any
    ) -> foast.FunctionDefinition:
        new_closure_vars: list[foast.Symbol] = []
        for sym in node.closure_vars:
            if not isinstance(self.closure_vars[sym.id], type):
                new_symbol: foast.Symbol = foast.Symbol(
                    id=sym.id,
                    location=sym.location,
                    type=type_translation.from_value(self.closure_vars[sym.id]),
                )
                new_closure_vars.append(new_symbol)
            else:
                new_closure_vars.append(sym)
        return foast.FunctionDefinition(
            id=node.id,
            params=node.params,
            body=node.body,
            closure_vars=new_closure_vars,
            type=node.type,
            location=node.location,
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import copy
import functools
from typing import Any, Optional

import dace
from dace import (
    dtypes as dace_dtypes,
    properties as dace_properties,
    subsets as dace_sbs,
    transformation as dace_transformation,
)
from dace.sdfg import nodes as dace_nodes, propagation as dace_propagation, utils as dace_sutils

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


@dace_properties.make_properties
class MoveDataflowIntoIfBody(dace_transformation.SingleStateTransformation):
    """The transformation moves dataflow into the if branches.

    Essentially transforms code from this
    ```python
    __arg1 = foo(...)
    __arg2 = bar(...)
    if __cond:
        __output = __arg1
    else:
        __output = __arg2
    ```
    into this
    ```python
    if __cond:
        __output = foo(...)
    else:
        __output = bar(...)
    ```
    I.e. computation that is needed in only one branch is relocated or inlined into
    that specific branch, which might reduce computation.

    Despite its name the transformation is not only able to handle `if` expressions,
    but more general `switch` like expressions. The requirements are:
    - The expression must be inside a nested SDFG containing only a `ConditionalBlock`.
    - The matched nested SDFG must be inside a Map (might be dropped).
    - For every incoming connector there must be exactly one AccessNode in the
        entire nested SDFG (might be dropped).
    - Every branch must write to every output of the nested SDFG.
    - The only dataflow allowed inside the branches is `(i) -> (o)` i.e. AccessNode
        to AccessNode connections.

    Furthermore, the transformation should be applied as long as possible, i.e.
    it should be applied by passing it to `SDFG.apply_transformations_repeated()`.

    Args:
        ignore_upstream_blocks: If `True` do not require that upstream `if_block`s
            have to be processed first.

    Note:
        - If there is a chain of suitable `if` expression, i.e. an `if` expression is
            in the upstream dataflow of the one that is currently matched, then the
            transformation will not apply. This is done to ensure that everything is
            properly inlined into the branches. This behaviour can be disabled by
            setting `ignore_upstream_blocks` to `True`.
            The reason for this behaviour is that the current implementation can only
            handle the case where the `if` expression is _directly_ inside a Map.

    Todo:
        - Allow that an inconnector can be used multiple times, as long as it is in
            different branches.
        - Extend the implementation that it is also able to handle `if_blocks` that
            are not inside a Map. This would allow to drop the need to process
            upstream `if` expressions first.
    """

    if_block = dace_transformation.PatternNode(dace_nodes.NestedSDFG)

    ignore_upstream_blocks = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="If 'True' ignore 'if_block's that are upstream.",
    )

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.if_block)]

    def __init__(
        self,
        ignore_upstream_blocks: Optional[bool] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if ignore_upstream_blocks is not None:
            self.ignore_upstream_blocks = ignore_upstream_blocks
        super().__init__(*args, **kwargs)

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        scope_dict = graph.scope_dict()
        if_block: dace_nodes.NestedSDFG = self.if_block
        assert isinstance(if_block, dace_nodes.NestedSDFG)

        # NOTE: The main benefit of requiring that the `if_block` must be inside a
        #   Map is that we know that every transient we encounter is single use data.
        enclosing_map = scope_dict[if_block]
        if not isinstance(enclosing_map, dace_nodes.MapEntry):
            return False

        # Test if the `if_block` is valid. This will also give us the names.
        if_block_spec = self._partition_if_block(if_block)
        if if_block_spec is None:
            return False

        # Compute the dataflow that is relocated.
        raw_relocatable_dataflow, non_relocatable_dataflow = (
            {
                conn_name: gtx_transformations.utils.find_upstream_nodes(
                    start=if_block,
                    state=graph,
                    start_connector=conn_name,
                    limit_node=enclosing_map,
                )
                for conn_name in conn_names
            }
            for conn_names in if_block_spec
        )
        relocatable_dataflow = self._filter_relocatable_dataflow(
            sdfg=sdfg,
            state=graph,
            if_block=if_block,
            raw_relocatable_dataflow=raw_relocatable_dataflow,
            non_relocatable_dataflow=non_relocatable_dataflow,
        )

        # If no branch has something to inline then we are done.
        if all(len(rel_df) == 0 for rel_df in relocatable_dataflow.values()):
            return False

        # Because the transformation can only handle `if` expressions that
        #  are _directly_ inside a Map, we must check if the upstream contains
        #  suitable `if` expressions that must be processed first. The simplest way
        #  is to return `False` and not apply. However, this implies that the
        #  transformation is applied in a loop until it applies nowhere anymore.
        # NOTE: This is a restriction due to the current implementation.
        if not self.ignore_upstream_blocks:
            for reloc_dataflow in relocatable_dataflow.values():
                if any(
                    self._has_if_block_relocatable_dataflow(
                        sdfg=sdfg,
                        state=graph,
                        upstream_if_block=upstream_if_block,
                        enclosing_map=enclosing_map,
                    )
                    for upstream_if_block in reloc_dataflow
                    if isinstance(upstream_if_block, dace_nodes.NestedSDFG)
                ):
                    return False

        return True

    def apply(
        self,
        graph: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        if_block: dace_nodes.NestedSDFG = self.if_block
        if_block_spec = self._partition_if_block(if_block)
        assert if_block_spec is not None
        enclosing_map = graph.scope_dict()[if_block]

        # Find the dataflow that should be relocated.
        raw_relocatable_dataflow, non_relocatable_dataflow = (
            {
                conn_name: gtx_transformations.utils.find_upstream_nodes(
                    start=if_block,
                    state=graph,
                    start_connector=conn_name,
                    limit_node=enclosing_map,
                )
                for conn_name in conn_names
            }
            for conn_names in if_block_spec
        )
        relocatable_dataflow = self._filter_relocatable_dataflow(
            sdfg=sdfg,
            state=graph,
            if_block=if_block,
            raw_relocatable_dataflow=raw_relocatable_dataflow,
            non_relocatable_dataflow=non_relocatable_dataflow,
        )

        # Finally relocate the dataflow
        for conn_name, nodes_to_move in relocatable_dataflow.items():
            self._replicate_dataflow_into_branch(
                state=graph,
                sdfg=sdfg,
                if_block=if_block,
                enclosing_map=enclosing_map,
                nodes_to_move=nodes_to_move,
                connector=conn_name,
            )

        self._update_symbol_mapping(if_block, sdfg)

        self._remove_outside_dataflow(
            sdfg=sdfg,
            state=graph,
            relocatable_dataflow=relocatable_dataflow,
        )

        # Because we relocate some node it seems that DaCe gets a bit confused.
        #  So we have to reset the list. Without it the `test_if_mover_chain`
        #  test would fail.
        sdfg.reset_cfg_list()

        # Readjust the Subsets.
        #  As a side effect this call will also properly propagate the nested SDFG `if_block`.
        dace_propagation.propagate_memlets_map_scope(sdfg, graph, enclosing_map)

    def _replicate_dataflow_into_branch(
        self,
        sdfg: dace.SDFG,
        state: dace.SDFGState,
        if_block: dace_nodes.NestedSDFG,
        enclosing_map: dace_nodes.MapEntry,
        nodes_to_move: set[dace_nodes.Node],
        connector: str,
    ) -> None:
        """Replicate the dataflow in `nodes_to_move` from `state` into `if_block`.

        First the function will determine into which branch, inside `if_block`,
        the dataflow has to be replicated. It will then copy the dataflow, nodes
        listed in `nodes_to_move` and insert them into that state.
        The function will then create the edges to connect them in the same way
        as they where outside. If there is an outer data dependency, for example
        a read to a global memory, then the function will patch that inside the
        `if_block`.
        At the end the function will remove the `connector`, but it will not remove
        the original dataflow.

        Args:
            sdfg: The sdfg that we process, the one that contains `state`.
            state: The state we operate on, the one that contains `if_block`.
            if_block: The `if_block` into which we inline.
            enclosing_map: The enclosing map.
            nodes_to_move: The list of nodes that should be removed.
            connector: The connector that should be inlined.
        """
        # Nothing to relocate nothing to do.
        if len(nodes_to_move) == 0:
            return

        inner_sdfg: dace.SDFG = if_block.sdfg
        branch_state, connector_node = self._find_branch_for(
            if_block=if_block,
            connector=connector,
        )

        # There might be AccessNodes inside `nodes_to_move`, we now have to make sure
        #  that they are present inside the nested ones. By our base assumption they
        #  are transients, because they are only used in one place
        for node in nodes_to_move:
            if not isinstance(node, dace_nodes.AccessNode):
                continue
            assert node.data not in inner_sdfg.arrays
            assert sdfg.arrays[node.data].transient
            # TODO(phimuell): Handle the case we need to rename something.
            inner_sdfg.add_datadesc(
                node.data,
                sdfg.arrays[node.data].clone(),
                find_new_name=False,
            )

        # Replicate the nodes. Also make a mapping that allows to map the old ones
        #  to the new ones.
        new_nodes: dict[dace_nodes.Node, dace_nodes.Node] = {
            old_node: copy.deepcopy(old_node) for old_node in nodes_to_move
        }
        branch_state.add_nodes_from(new_nodes.values())

        # Now add the edges between the edges that have been replicated inside the
        #  branch state, these are the outgoing edges. The data dependencies of the
        #  nodes that were not relocated are still missing.
        for node in nodes_to_move:
            for oedge in state.out_edges(node):
                if oedge.dst is if_block:
                    assert oedge.dst_conn == connector
                    # TODO(phimuell): Make subsets complete.
                    branch_state.add_edge(
                        new_nodes[oedge.src],
                        oedge.src_conn,
                        connector_node,
                        None,
                        dace.Memlet.from_memlet(oedge.data),
                    )
                else:
                    assert oedge.dst in nodes_to_move
                    branch_state.add_edge(
                        new_nodes[oedge.src],
                        oedge.src_conn,
                        new_nodes[oedge.dst],
                        oedge.dst_conn,
                        dace.Memlet.from_memlet(oedge.data),
                    )

        # Now we have to satisfy the data dependencies, i.e. forward all nodes that
        #  could not have been moved inside `if_block` but are still needed to compute
        #  the final result. We find them by scanning the input edges of the nodes
        #  that have been relocated.
        for node in nodes_to_move:
            for iedge in state.in_edges(node):
                if iedge.src in nodes_to_move:
                    # Inner data dependency, there is nothing to do and the edge was
                    #  created above.
                    continue
                if iedge.data.is_empty():
                    # Empty Memlets are there to maintain some order relation, "happens
                    #  before". Depending on the situation we can remove or have to
                    #  recreate them. The case where the connection comes from a
                    #  node within the relocated dataflow is handled above.
                    assert iedge.src is enclosing_map
                    continue

                # Now we have to figuring out where the data is coming from, since
                #  the data was not relocated, we must patch it into `if_block`.
                if iedge.src is enclosing_map:
                    # The data is coming from outside the Map scope, i.e. not defined
                    #  inside the Map scope, so we have to trace it back.
                    memlet_path = state.memlet_path(iedge)
                    outer_data = memlet_path[0].src
                else:
                    # The data is defined somewhere in the Map scope itself.
                    outer_data = iedge.src
                # TODO(phimuell): It is possible that this does not lead to an
                #   AccessNode on the outside, but to something inside the Map scope
                #   such as the MapExit of an inner map. To handle such a case we need
                #   to construct the set of nodes to move differently, i.e.
                #   considering this case already there.
                if not isinstance(outer_data, dace_nodes.AccessNode):
                    raise NotImplementedError()
                assert not gtx_transformations.utils.is_view(outer_data, sdfg)

                # If the data is not yet available in the inner SDFG made
                #  patch it through.
                if outer_data.data not in inner_sdfg.arrays:
                    inner_desc = sdfg.arrays[outer_data.data].clone()
                    inner_desc.transient = False
                    # TODO(phimuell): Handle the case we need to rename something.
                    inner_sdfg.add_datadesc(outer_data.data, inner_desc, False)
                    # TODO(phimeull): We pass the whole data inside the SDFG.
                    #   Find out if there are cases where this is wrong.
                    state.add_edge(
                        iedge.src,
                        iedge.src_conn,
                        if_block,
                        outer_data.data,
                        dace.Memlet(
                            data=outer_data.data, subset=dace_sbs.Range.from_array(inner_desc)
                        ),
                    )
                    if_block.add_in_connector(outer_data.data)
                else:
                    # This is the case that we found a node, that refers to data that
                    #  was already patched into the `if_block`. We would have to remove
                    #  this, but since this function just replicates the dataflow,
                    #  it will not do that. Instead we postpone this to the cleanup
                    #  phase, see `_remove_outside_dataflow()`.
                    pass

                if outer_data not in new_nodes:
                    assert all(
                        outer_data.data != mapped_node.data
                        for mapped_node in new_nodes.values()
                        if isinstance(mapped_node, dace_nodes.AccessNode)
                    )
                    assert outer_data.data in inner_sdfg.arrays
                    assert not inner_sdfg.arrays[outer_data.data].transient
                    new_nodes[outer_data] = branch_state.add_access(
                        outer_data.data, copy.copy(outer_data.debuginfo)
                    )

                # Now create the edge in the inner state.
                branch_state.add_edge(
                    new_nodes[outer_data],
                    None,
                    new_nodes[iedge.dst],
                    iedge.dst_conn,
                    copy.deepcopy(iedge.data),
                )

        # The old connector name is no longer valid.
        inner_sdfg.arrays[connector].transient = True
        if_block.remove_in_connector(connector)

    def _remove_outside_dataflow(
        self,
        sdfg: dace.SDFG,
        state: dace.SDFGState,
        relocatable_dataflow: dict[str, set[dace_nodes.Node]],
    ) -> None:
        """Removes the original dataflow, that has been relocated.

        The function will also remove data containers that are no longer in use.
        """
        all_relocatable_dataflow: set[dace_nodes.Node] = functools.reduce(
            lambda s1, s2: s1.union(s2), relocatable_dataflow.values(), set()
        )

        # Before we can clean the original nodes, we must clean the dataflow. If a
        #  node, that was relocated, has incoming connections we must remove them
        #  and the parent dataflow.
        for node_to_remove in all_relocatable_dataflow:
            for iedge in list(state.in_edges(node_to_remove)):
                if iedge.src in all_relocatable_dataflow:
                    continue
                dace_sutils.remove_edge_and_dangling_path(state, iedge)

            if isinstance(node_to_remove, dace_nodes.AccessNode):
                assert node_to_remove.desc(sdfg).transient
                sdfg.remove_data(node_to_remove.data, validate=False)

        # Remove the original nodes (data descriptors were deleted in the loop above).
        state.remove_nodes_from(all_relocatable_dataflow)

    def _update_symbol_mapping(
        self,
        if_block: dace_nodes.NestedSDFG,
        parent: dace.SDFG,
    ) -> None:
        """Updates the symbol mapping of the nested SDFG.

        The function assumes that the symbols that are missing in the nested SDFG
        are available in the parent SDFG.
        """
        symbol_mapping = if_block.symbol_mapping
        missing_symbols = [ms for ms in if_block.sdfg.free_symbols if ms not in symbol_mapping]
        symbol_mapping.update({s: s for s in missing_symbols})
        if_block.symbol_mapping = symbol_mapping  # Performs conversion.

        # Add new global symbols to nested SDFG.
        #  The code is based on `SDFGState.add_nested_sdfg()`.
        if_block_symbols = if_block.sdfg.symbols
        parent_symbols = parent.symbols
        for new_sym in missing_symbols:
            if new_sym in if_block_symbols:
                # The symbol is already known, so we check that it is the same type as in the
                #  parent SDFG.
                assert if_block_symbols[new_sym] == parent_symbols[new_sym]

            elif new_sym in parent_symbols:
                # The symbol is known to the parent SDFG, so take the type from there.
                if_block.sdfg.add_symbol(new_sym, parent_symbols[new_sym])

            else:
                # Figuring out the type of the symbol based on the computation we do.
                # TODO(phimuell): Maybe switch to `symbols_defined_at()` as it is indicated
                #   in the `SDFGState.add_nested_sdfg()` function.
                if_block.sdfg.add_symbol(
                    new_sym,
                    dace.codegen.tools.type_inference.infer_expr_type(new_sym, parent_symbols)
                    or dace_dtypes.typeclass(int),
                )

    def _find_branch_for(
        self,
        if_block: dace_nodes.NestedSDFG,
        connector: str,
    ) -> tuple[dace.SDFGState, dace_nodes.AccessNode]:
        """
        Locates the branch and the AccessNode to where the dataflow should be relocated.
        """
        inner_sdfg: dace.SDFG = if_block.sdfg
        conditional_block: dace.sdfg.state.ConditionalBlock = next(iter(inner_sdfg.nodes()))

        # This will locate the state where the first AccessNode that refers to
        #  `connector` is found. Since `_partition_if_block()` makes sure that
        #  there is only one match this is okay. But it must be changed, if we
        #  lift this restriction.
        for inner_state in conditional_block.all_states():
            connector_nodes: list[dace_nodes.AccessNode] = [
                dnode for dnode in inner_state.data_nodes() if dnode.data == connector
            ]
            if len(connector_nodes) == 0:
                continue
            break
        else:
            raise ValueError(f"Did not find a branch associated to '{connector}'.")

        assert isinstance(inner_state, dace.SDFGState)
        assert inner_state.in_degree(connector_nodes[0]) == 0
        assert inner_state.out_degree(connector_nodes[0]) > 0
        return inner_state, connector_nodes[0]

    def _has_if_block_relocatable_dataflow(
        self,
        sdfg: dace.SDFG,
        state: dace.SDFGState,
        upstream_if_block: dace_nodes.NestedSDFG,
        enclosing_map: dace_nodes.MapEntry,
    ) -> bool:
        """Check if `upstream_if_block` has relocatable dataflow.

        The function is used to enforce the rule that no `if` expression should be
        processed before all suitable `if` expressions in its upstream dataflow of
        the matched `if_block` has been processed first.

        Args:
            sdfg: The SDFG on which we operate.
            state: The state on which we operate on.
            upstream_if_block: The `if` expression that was found in the relocatable
                dataflow of the matched `if_block`.
            enclosing_map: The limiting node, i.e. the MapEntry of the Map `if_block`
                is located in.
        """
        if_block_spec = self._partition_if_block(upstream_if_block)
        if if_block_spec is None:
            return False

        raw_relocatable_dataflow, non_relocatable_dataflow = (
            {
                conn_name: gtx_transformations.utils.find_upstream_nodes(
                    start=upstream_if_block,
                    state=state,
                    start_connector=conn_name,
                    limit_node=enclosing_map,
                )
                for conn_name in conn_names
            }
            for conn_names in if_block_spec
        )
        filtered_relocatable_dataflow = self._filter_relocatable_dataflow(
            sdfg=sdfg,
            state=state,
            if_block=upstream_if_block,
            raw_relocatable_dataflow=raw_relocatable_dataflow,
            non_relocatable_dataflow=non_relocatable_dataflow,
        )
        if all(len(rel_df) == 0 for rel_df in filtered_relocatable_dataflow.values()):
            return False

        return True

    def _filter_relocatable_dataflow(
        self,
        sdfg: dace.SDFG,
        state: dace.SDFGState,
        if_block: dace_nodes.NestedSDFG,
        raw_relocatable_dataflow: dict[str, set[dace_nodes.Node]],
        non_relocatable_dataflow: dict[str, set[dace_nodes.Node]],
    ) -> dict[str, set[dace_nodes.Node]]:
        """Partition the dependencies.

        The function expects the dataflow that is upstream of every connector
        of the `if_block`. The function will then scan the dataflow and compute
        the parts that actually can be relocated and returns a `dict` mapping
        every relocatable input connector to the set of nodes that can be relocated.

        Note that the sets that are returned by this function are distinct.

        Args:
            state: The state on which we operate.
            if_block: The `if_block` that is processed.
            raw_relocatable_dataflow: The connectors and their associated dataflow
                that can be relocated, not yet filtered.
            non_relocatable_dataflow: The connectors and their associated dataflow
                that can not be relocated.
        """

        # Remove the parts of the dataflow that is unrelocatable.
        all_non_relocatable_dataflow: set[dace_nodes.Node] = functools.reduce(
            lambda s1, s2: s1.union(s2), non_relocatable_dataflow.values(), set()
        )
        relocatable_dataflow = {
            conn_name: rel_df.difference(all_non_relocatable_dataflow)
            for conn_name, rel_df in raw_relocatable_dataflow.items()
        }

        # Now we determine the nodes that are in more than one sets.
        #  These sets must be removed, from the individual sets.
        known_nodes: set[dace_nodes.Node] = set()
        multiple_df_nodes: set[dace_nodes.Node] = set()
        for rel_df in relocatable_dataflow.values():
            seen_before: set[dace_nodes.Node] = known_nodes.intersection(rel_df)
            if len(seen_before) != 0:
                multiple_df_nodes.update(seen_before)
            known_nodes.update(rel_df)
        relocatable_dataflow = {
            conn_name: rel_df.difference(multiple_df_nodes)
            for conn_name, rel_df in relocatable_dataflow.items()
        }

        # However, not all dataflow can be moved inside the branch. For example if
        #  something is used outside the dataflow, that is moved inside the `if`,
        #  then we can not relocate it.
        # TODO(phimuell): If we operate outside of a Map we also have to make sure that
        #   the data is single use data, is not an AccessNode that refers to global
        #   memory nor is a source AccessNode.
        def filter_nodes(
            branch_nodes: set[dace_nodes.Node],
            sdfg: dace.SDFG,
            state: dace.SDFGState,
        ) -> set[dace_nodes.Node]:
            # For this to work the `if_block` must be considered part, we remove it later.
            branch_nodes.add(if_block)
            has_been_updated = True
            while has_been_updated:
                has_been_updated = False
                for node in list(branch_nodes):
                    if node is if_block:
                        continue
                    if any(oedge.dst not in branch_nodes for oedge in state.out_edges(node)):
                        branch_nodes.remove(node)
                        has_been_updated = True
            assert if_block in branch_nodes
            branch_nodes.remove(if_block)
            return branch_nodes

        return {
            conn_name: filter_nodes(rel_df, sdfg, state)
            for conn_name, rel_df in relocatable_dataflow.items()
        }

    def _partition_if_block(
        self,
        if_block: dace_nodes.NestedSDFG,
    ) -> Optional[tuple[set[str], set[str]]]:
        """Check if `if_block` can be processed and partition the input connectors.

        The function will check if `if_block` has the right structure, i.e. if it is
        roughly equivalent to the Python expression `a if cond else b`.
        It will also identify which input connectors refer to dataflow that can
        be inlined into the `if_block` and which can not.

        Returns:
            If `if_block` is unsuitable the function will return `None`.
            If `if_block` meets the structural requirements the function will return
            two sets of strings. The first set contains the connectors that can be
            relocated and the second one of the conditions that can not be relocated.
        """
        # There shall only be one output and three inputs with given names.
        if len(if_block.out_connectors.keys()) == 0:
            return None

        # These are all the output names.
        output_names: set[str] = set(if_block.out_connectors.keys())

        # We require that the nested SDFG contains a single node, which is a
        #  `ConditionalBlock` containing two branches.
        inner_sdfg: dace.SDFG = if_block.sdfg
        if inner_sdfg.number_of_nodes() != 1:
            return None
        inner_if_block = next(iter(inner_sdfg.nodes()))
        if not isinstance(inner_if_block, dace.sdfg.state.ConditionalBlock):
            return None

        # Defining it outside will ensure that there is only one AccessNode for every
        #  inconnector, which is something `_find_branch_for()` relies on.
        reference_count: dict[str, int] = {conn_name: 0 for conn_name in if_block.in_connectors}

        for _, branch in inner_if_block.branches:
            output_count: dict[str, int] = {conn_name: 0 for conn_name in output_names}
            for inner_state in branch.all_states():
                assert isinstance(inner_state, dace.SDFGState)
                for node in inner_state.nodes():
                    if not isinstance(node, dace_nodes.AccessNode):
                        return None
                    if node.data in reference_count:
                        reference_count[node.data] += 1
                        exp_in_deg, exp_out_deg = 0, 1
                    elif node.data in output_count:
                        output_count[node.data] += 1
                        exp_in_deg, exp_out_deg = 1, 0
                    else:
                        return None
                    if inner_state.in_degree(node) != exp_in_deg:
                        return None
                    if inner_state.out_degree(node) != exp_out_deg:
                        return None
            # Each branch must write to all outputs.
            # TODO(phimuell): Think if this should be lifted.
            if any(count != 1 for count in output_count.values()):
                return None

        # The connectors that can be pulled inside must appear exactly once.
        #  In theory they could appear more, but then we would have to replicate
        #  the dataflow to different locations which is not supported.
        #  So the ones that can be relocated were found exactly once. Zero would
        #  mean they can not be relocated and more than one means that we do not
        #  support it yet.
        relocatable_connectors = {
            conn_name for conn_name, conn_count in reference_count.items() if conn_count == 1
        }
        non_relocatable_connectors = {
            conn_name
            for conn_name in reference_count.keys()
            if conn_name not in relocatable_connectors
        }
        if len(non_relocatable_connectors) == 0:
            return None
        if len(relocatable_connectors) == 0:
            return None
        return relocatable_connectors, non_relocatable_connectors



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
from typing import Optional, Sequence, TypeAlias

import dace
import sympy
from dace import subsets as dace_subsets

from gt4py.next import common as gtx_common
from gt4py.next.iterator import ir as gtir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, domain_utils
from gt4py.next.program_processors.runners.dace import gtir_to_sdfg_utils


@dataclasses.dataclass(frozen=True)
class FieldopDomainRange:
    """
    Represents the range of a field operator domain in one dimension.

    It contains 3 elements:
        dim: dimension definition
        start: symbolic expression for lower bound (inclusive)
        stop: symbolic expression for upper bound (exclusive)
    """

    dim: gtx_common.Dimension
    start: dace.symbolic.SymbolicType
    stop: dace.symbolic.SymbolicType


FieldopDomain: TypeAlias = list[FieldopDomainRange]
"""Domain of a field operator represented as a list of `FieldopDomainRange` for each dimension."""


def extract_domain(node: gtir.Expr) -> FieldopDomain:
    """
    Visits the domain of a field operator and returns a list of dimensions and
    the corresponding lower and upper bounds. The returned lower bound is inclusive,
    the upper bound is exclusive: [lower_bound, upper_bound[
    """

    domain = []

    if cpm.is_call_to(node, ("cartesian_domain", "unstructured_domain")):
        for named_range in node.args:
            assert cpm.is_call_to(named_range, "named_range")
            assert len(named_range.args) == 3
            axis = named_range.args[0]
            assert isinstance(axis, gtir.AxisLiteral)
            lower_bound, upper_bound = (
                gtir_to_sdfg_utils.get_symbolic(arg) for arg in named_range.args[1:3]
            )
            dim = gtx_common.Dimension(axis.value, axis.kind)
            domain.append(FieldopDomainRange(dim, lower_bound, upper_bound))

    elif isinstance(node, domain_utils.SymbolicDomain):
        for dim, drange in node.ranges.items():
            domain.append(
                FieldopDomainRange(
                    dim,
                    gtir_to_sdfg_utils.get_symbolic(drange.start),
                    gtir_to_sdfg_utils.get_symbolic(drange.stop),
                )
            )

    else:
        raise ValueError(f"Invalid domain {node}.")

    return domain


def simplify_domain_expr(expr: sympy.Basic, domain: FieldopDomain) -> dace.symbolic.SymbolicType:
    """Simplifies a symbolic expression by applying constraints from domain range.

    Dace uses sympy for symbolic expressions in the SDFG. By applying assumptions
    on a sympy expression, we may obtain a simplified expression.
    This is particularly important in the lowering of concat_where domain expressions,
    because it usually results in cleaner memlet subsets and better map fusion.

    Args:
        expr: The symbolic expression to simplify.
    Returns:
        A new symbolic expression.
    """
    for dim_range in domain:
        # We want to enforce the constraint `ub = lb + size`. The actual constraint
        # is given by the assumption that the `size` variable is integer and non-negative.
        size = sympy.var(f"__gtir_{dim_range.dim.value}_size", integer=True, negative=False)
        expr = expr.subs(dim_range.start, dim_range.stop - size).subs(
            size, dim_range.stop - dim_range.start
        )
    return dace.symbolic.simplify_ext(expr)


def get_domain_indices(
    dims: Sequence[gtx_common.Dimension], origin: Optional[Sequence[dace.symbolic.SymExpr]]
) -> dace_subsets.Indices:
    """
    Helper function to construct the list of indices for a field domain, applying
    an optional origin in each dimension as start index.

    Args:
        dims: The field dimensions.
        origin: The domain start index in each dimension. If set to `None`, assume all zeros.

    Returns:
        A list of indices for field access in dace arrays. As this list is returned
        as `dace.subsets.Indices`, it should be converted to `dace.subsets.Range` before
        being used in memlet subset because ranges are better supported throughout DaCe.
    """
    assert len(dims) != 0
    index_variables = [
        dace.symbolic.pystr_to_symbolic(gtir_to_sdfg_utils.get_map_variable(dim)) for dim in dims
    ]
    origin = [0] * len(index_variables) if origin is None else origin
    return dace_subsets.Indices(
        [index - start_index for index, start_index in zip(index_variables, origin, strict=True)]
    )


def get_field_layout(
    field_domain: FieldopDomain,
    target_domain: FieldopDomain,
) -> tuple[list[gtx_common.Dimension], list[dace.symbolic.SymExpr], list[dace.symbolic.SymExpr]]:
    """
    Parse the field operator domain and generate the shape of the result field.

    Note that this function also ensures that the array shape computed from the
    domain range is non-negative. A negative shape can occur in concat_where
    expressions, where it can happen that 'stop' value is smaller than 'start'.
    Also note that this _strange_ domain with 'start' > 'stop' is usually propagated
    to the input arguments of a concat_where expression (the child nodes), thus
    the lowering of regular field operators also needs to apply the sanity check
    in order to avoid allocation of temporary fields with negative size.

    Args:
        field_domain: The field operator domain.
        target_domain: Domain of the target field in the root `SetAt` expression.

    Returns:
        A tuple of three lists containing:
            - the domain dimensions
            - the domain origin, that is the start indices in all dimensions
            - the domain size in each dimension
    """
    if len(field_domain) == 0:
        return [], [], []
    domain_dims = [domain_range.dim for domain_range in field_domain]
    domain_origin = [domain_range.start for domain_range in field_domain]
    domain_shape = [
        simplify_domain_expr(sympy.Max(0, (domain_range.stop - domain_range.start)), target_domain)
        for domain_range in field_domain
    ]
    return domain_dims, domain_origin, domain_shape



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import functools
from typing import Any, Callable, Dict, List


# TODO test


class _fun_dispatcher:
    def __init__(self, dispatcher, fun) -> None:
        self.dispatcher = dispatcher
        self.fun = fun
        functools.update_wrapper(self, fun)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self.dispatcher.key is None:
            return self.fun(*args, **kwargs)
        else:
            return self.dispatcher._funs[self.dispatcher.key][self.fun.__name__](*args, **kwargs)

    def register(self, key):
        self.dispatcher.register_key(key)

        def _impl(fun):
            self.dispatcher._funs[key][self.fun.__name__] = fun
            return fun

        # required for direct call to dispatched functions (see roundtrip).
        return _impl


class Dispatcher:
    def __init__(self) -> None:
        self._funs: Dict[str, Dict[str, Callable]] = {}
        self.key_stack: List[str] = []

    @property
    def key(self):
        return self.key_stack[-1] if self.key_stack else None

    def register_key(self, key):
        if key not in self._funs:
            self._funs[key] = {}

    def push_key(self, key):
        if key not in self._funs:
            raise RuntimeError(f"Key '{key}' not registered.")
        self.key_stack.append(key)

    def pop_key(self):
        self.key_stack.pop()

    def clear_key(self):
        self.key_stack = []

    def __call__(self, fun):
        return self.dispatch(fun)

    def dispatch(self, fun):
        return _fun_dispatcher(self, fun)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.next.iterator.transforms.concat_where.canonicalize_domain_argument import (
    canonicalize_domain_argument,
)
from gt4py.next.iterator.transforms.concat_where.expand_tuple_args import expand_tuple_args
from gt4py.next.iterator.transforms.concat_where.transform_to_as_fieldop import (
    transform_to_as_fieldop,
)


__all__ = ["canonicalize_domain_argument", "expand_tuple_args", "transform_to_as_fieldop"]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Implementation of stencil functions defined with symbolic Python functions.

This module contains functions to generate callable objects implementing
a high-level stencil function definition using a specific code generating backend.
"""

from __future__ import annotations

import types
from typing import TYPE_CHECKING, Any

from gt4py.cartesian import frontend as gt_frontend
from gt4py.cartesian.stencil_builder import StencilBuilder
from gt4py.cartesian.type_hints import StencilFunc


if TYPE_CHECKING:
    from gt4py.cartesian.definitions import BuildOptions
    from gt4py.cartesian.stencil_object import StencilObject


def load_stencil(
    frontend_name: str,
    backend_name: str,
    definition_func: StencilFunc,
    externals: dict[str, Any],
    dtypes: dict[type, type],
    build_options: BuildOptions,
) -> type[StencilObject]:
    """Generate a new class object implementing the provided definition."""
    # Load components
    frontend = gt_frontend.from_name(frontend_name)

    builder = (
        StencilBuilder(
            definition_func, options=build_options, backend=backend_name, frontend=frontend
        )
        .with_externals(externals)
        .with_dtypes(dtypes)
    )

    return builder.build()


def gtscript_loader(
    definition_func: StencilFunc,
    backend: str,
    build_options: BuildOptions,
    externals: dict[str, Any],
    dtypes: dict[type, type],
) -> StencilObject:
    if not isinstance(definition_func, types.FunctionType):
        raise ValueError("Invalid stencil definition object ({obj})".format(obj=definition_func))

    if not build_options.name:
        build_options.name = f"{definition_func.__name__}"
    stencil_class = load_stencil(
        "gtscript", backend, definition_func, externals, dtypes, build_options
    )

    return stencil_class()



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import Any, Dict, Optional, Union

import dace
from dace import (
    properties as dace_properties,
    subsets as dace_subsets,
    transformation as dace_transformation,
)
from dace.sdfg import graph as dace_graph, nodes as dace_nodes, propagation as dace_propagation
from dace.transformation.dataflow import map_fusion_helper as dace_mfhelper
from dace.transformation.passes import analysis as dace_analysis

from gt4py.next.program_processors.runners.dace import (
    transformations as gtx_transformations,
    utils as gtx_dace_utils,
)
from gt4py.next.program_processors.runners.dace.transformations import (
    map_fusion_utils as gtx_mfutils,
)


def gt_horizontal_map_split_fusion(
    sdfg: dace.SDFG,
    run_simplify: bool,
    fuse_possible_maps: bool,
    consolidate_edges_only_if_not_extending: bool,
    skip: Optional[set[str]] = None,
    validate: bool = True,
    validate_all: bool = False,
) -> int:
    transformations = [
        HorizontalSplitMapRange(
            fuse_possible_maps=fuse_possible_maps,
            only_toplevel_maps=True,
            consolidate_edges_only_if_not_extending=consolidate_edges_only_if_not_extending,
        ),
        gtx_transformations.MapFusionHorizontal(
            only_if_common_ancestor=True,
            only_inner_maps=False,
            only_toplevel_maps=True,
            consolidate_edges_only_if_not_extending=consolidate_edges_only_if_not_extending,
        ),
    ]

    ret = sdfg.apply_transformations_repeated(
        transformations,
        validate=False,
        validate_all=validate_all,
    )

    if run_simplify:
        if skip is None:
            skip = gtx_transformations.constants.GT_SIMPLIFY_DEFAULT_SKIP_SET
        if not consolidate_edges_only_if_not_extending:
            skip = skip.union(["ConsolidateEdges"])
        gtx_transformations.gt_simplify(
            sdfg=sdfg,
            validate=False,
            validate_all=validate_all,
            skip=skip,
        )
    elif validate and (not validate_all):
        sdfg.validate()

    return ret


def gt_vertical_map_split_fusion(
    sdfg: dace.SDFG,
    run_simplify: bool,
    consolidate_edges_only_if_not_extending: bool,
    single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
    skip: Optional[set[str]] = None,
    validate: bool = True,
    validate_all: bool = False,
) -> int:
    if single_use_data is None:
        find_single_use_data = dace_analysis.FindSingleUseData()
        single_use_data = find_single_use_data.apply_pass(sdfg, None)

    # TODO: Restrict MapFusion such that it only applies to the Maps that have
    #   been split and not some other random Maps.
    transformations = [
        VerticalSplitMapRange(
            only_toplevel_maps=True,
        ),
        gtx_transformations.SplitAccessNode(single_use_data=single_use_data),
        gtx_transformations.MapFusionVertical(
            only_inner_maps=False,
            only_toplevel_maps=True,
            consolidate_edges_only_if_not_extending=consolidate_edges_only_if_not_extending,
        ),
    ]
    # TODO(phimuell): Remove that hack once [issue#1911](https://github.com/spcl/dace/issues/1911)
    #   has been solved.
    transformations[-1]._single_use_data = single_use_data  # type: ignore[attr-defined]

    ret = sdfg.apply_transformations_repeated(
        transformations,
        validate=False,
        validate_all=validate_all,
    )

    if run_simplify:
        if skip is None:
            skip = gtx_transformations.constants.GT_SIMPLIFY_DEFAULT_SKIP_SET
        if not consolidate_edges_only_if_not_extending:
            skip = skip.union(["ConsolidateEdges"])
        gtx_transformations.gt_simplify(
            sdfg=sdfg,
            validate=validate,
            validate_all=validate_all,
            skip=skip,
        )
    elif validate and (not validate_all):
        sdfg.validate()

    return ret


@dace_properties.make_properties
class SplitMapRange(dace_transformation.SingleStateTransformation):
    """
    Identify overlapping range between maps, and split the maps into
    maps with smaller ranges in order to create two maps with common
    range that can be fused together using the `HorizontalSplitMapRange`
    or `VerticalSplitMapRange` transformations.
    For example, given a map with range [0:10, 14:80] and a second map
    with range [5:15, 0:80], this transformation will split the first map
    into two maps with ranges [0:5, 14:80] and [5:10, 14:80], and the second
    map into four maps with ranges [0:10, 0:14], [0:10, 14:80] [10:15, 0:14]
    and [10:15, 14:80]. This will allow the `HorizontalSplitMapRange` or
    `VerticalSplitMapRange` transformations to fuse the maps together.
    """

    only_toplevel_maps = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="Only perform fusing if the Maps are in the top level.",
    )

    def __init__(
        self,
        only_toplevel_maps: Optional[bool] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        if only_toplevel_maps is not None:
            self.only_toplevel_maps = only_toplevel_maps

    def split_maps(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        sdfg: dace.SDFG,
        first_map_entry: dace_nodes.MapEntry,
        first_map_exit: dace_nodes.MapExit,
        second_map_entry: dace_nodes.MapEntry,
        second_map_exit: dace_nodes.MapExit,
    ) -> tuple[list[dace_nodes.MapEntry], list[dace_nodes.MapEntry]]:
        """Split the map range in order to obtain an overlapping range between the first and second map.

        The function returns a pair of lists. The first `list` contains the new `MapEntry`s that
        were generated by splitting the first Map and the second `list` contains the `MapEntry`s
        generated by splitting the second Map.
        """
        splitted_range = gtx_mfutils.split_overlapping_map_range(
            first_map_entry.map, second_map_entry.map
        )
        assert splitted_range is not None

        scope_dict = graph.scope_dict()
        assert scope_dict[first_map_entry] is scope_dict[second_map_entry]
        containing_scope_or_none = scope_dict[first_map_entry]

        first_map_splitted_range, second_map_splitted_range = splitted_range

        def _replace_ranged_map(
            map_entry: dace_nodes.MapEntry,
            map_exit: dace_nodes.MapExit,
            new_ranges: list[dace_subsets.Range],
        ) -> list[dace_nodes.MapEntry]:
            """Replace a map with multiple maps based on new_ranges."""
            new_map_entries: list[dace_nodes.MapEntry] = []
            for i, r in enumerate(new_ranges):
                new_map_entry, _ = gtx_mfutils.copy_map_graph_with_new_range(
                    sdfg, graph, map_entry, map_exit, r, str(i)
                )
                new_map_entries.append(new_map_entry)
            gtx_mfutils.delete_map(graph, map_entry, map_exit)  # remove the original first map
            return new_map_entries

        new_map_entries_from_first = _replace_ranged_map(
            first_map_entry,
            first_map_exit,
            first_map_splitted_range,
        )
        new_map_entries_from_second = _replace_ranged_map(
            second_map_entry,
            second_map_exit,
            second_map_splitted_range,
        )

        # Now we propagate the Memlets. If the original Maps where located at the global
        #  scope we have to propagate the maps from the new ones. However, if they are
        #  not at global scope we simply propagate the surrounding Map, since they are
        #  by assumption inside the same Map.
        if containing_scope_or_none is None:
            for new_map_entry in new_map_entries_from_first + new_map_entries_from_second:
                dace_propagation.propagate_memlets_map_scope(sdfg, graph, new_map_entry)
        else:
            dace_propagation.propagate_memlets_map_scope(sdfg, graph, containing_scope_or_none)

        # Workaround to ensure that some cache in DaCe has been cleared.
        # TODO(phimuell, iomaganaris): Before `hash_sdfg()` was used, but this was a
        #   very heavy operation. I think that this is enough.
        sdfg.reset_cfg_list()

        return new_map_entries_from_first, new_map_entries_from_second


@dace_properties.make_properties
class HorizontalSplitMapRange(SplitMapRange):
    """
    Identify overlapping range between parallel maps, and split the range in order
    to promote parallel map fusion.
    """

    first_map_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)
    second_map_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)

    fuse_possible_maps = dace_properties.Property(
        dtype=bool,
        default=True,
        desc="If `True`, the transformation will try to fuse maps that have the same range together.",
    )
    never_consolidate_edges = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="If `True`, always create a new connector, instead of reusing one that referring to the same data.",
    )
    consolidate_edges_only_if_not_extending = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="Only consolidate if this does not lead to an extension of the subset.",
    )

    def __init__(
        self,
        fuse_possible_maps: Optional[bool] = None,
        consolidate_edges_only_if_not_extending: Optional[bool] = None,
        never_consolidate_edges: Optional[bool] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        if fuse_possible_maps is not None:
            self.fuse_possible_maps = fuse_possible_maps
        if consolidate_edges_only_if_not_extending is not None:
            self.consolidate_edges_only_if_not_extending = consolidate_edges_only_if_not_extending
        if never_consolidate_edges is not None:
            self.never_consolidate_edges = never_consolidate_edges

    @classmethod
    def expressions(cls) -> Any:
        map_fusion_parallel_match = dace_graph.OrderedMultiDiConnectorGraph()
        map_fusion_parallel_match.add_nodes_from([cls.first_map_entry, cls.second_map_entry])
        return [map_fusion_parallel_match]

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        first_map_entry: dace_nodes.MapEntry = self.first_map_entry
        second_map_entry: dace_nodes.MapEntry = self.second_map_entry
        first_map: dace_nodes.Map = first_map_entry.map
        second_map: dace_nodes.Map = second_map_entry.map

        # Ensure that the Maps are in the same scope.
        scope_dict = graph.scope_dict()
        if scope_dict[self.first_map_entry] is not scope_dict[self.second_map_entry]:
            return False

        # Test if the map is in the right scope.
        map_scope: Union[dace_nodes.Node, None] = scope_dict[self.first_map_entry]
        if self.only_toplevel_maps and (map_scope is not None):
            return False

        first_map_src_data = {
            iedge.src.label
            for iedge in graph.in_edges(self.first_map_entry)
            if isinstance(iedge.src, dace_nodes.AccessNode)
        }
        second_map_src_data = {
            iedge.src.label
            for iedge in graph.in_edges(self.second_map_entry)
            if isinstance(iedge.src, dace_nodes.AccessNode)
        }

        # Test if the Maps are parallel.
        if not dace_mfhelper.is_parallel(
            graph=graph, node1=first_map_entry, node2=second_map_entry
        ):
            return False

        if len(first_map_src_data.intersection(second_map_src_data)) == 0:
            # no common source access node
            return False

        splitted_range = gtx_mfutils.split_overlapping_map_range(first_map, second_map)
        if splitted_range is None:
            return False

        return True

    def apply(
        self,
        graph: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        first_map_entry: dace_nodes.MapEntry = self.first_map_entry
        second_map_entry: dace_nodes.MapEntry = self.second_map_entry

        # Now split the first and second Map.
        first_map_fragments, second_map_fragments = self.split_maps(
            graph,
            sdfg,
            first_map_entry,
            graph.exit_node(first_map_entry),
            second_map_entry,
            graph.exit_node(second_map_entry),
        )

        # If we do not want to fuse the maps, we can stop here.
        if not self.fuse_possible_maps:
            return

        # Look which Maps can be fused together. It is important that a Map can be listed multiple
        #  times. However, by fusing we remove them from the state, which is a problem. We solve
        #  this by always using "first Maps" as `to_node`, thus they survive. Furthermore, we
        #  only assign a second Map once.
        # TODO(phimuell, iomaganaris): Check and improve that.
        matched_map_fragments: list[tuple[dace_nodes.MapEntry, dace_nodes.MapEntry]] = []
        unmatched_second_map_fragments = second_map_fragments.copy()
        for first_map_fragment in first_map_fragments:
            for unmatched_second_map_fragment in unmatched_second_map_fragments:
                if first_map_fragment.map.range == unmatched_second_map_fragment.map.range:
                    matched_map_fragments.append(
                        (first_map_fragment, unmatched_second_map_fragment)
                    )
                    unmatched_second_map_fragments.remove(unmatched_second_map_fragment)
                    break

        assert len(matched_map_fragments) > 0

        # We have to get the original scope dict before we start mutating the graph. Actually we
        #  would need to obtain the scope dict after every iteration again, which would require
        #  a rescan. But since the operations do not alter the scopes, at least not in a way that
        #  would affect us, we can be more efficient and get the thing at the beginning.
        scope_dict: Dict = graph.scope_dict()

        for first_map_fragment_entry, second_map_fragment_entry in matched_map_fragments:
            first_map_fragment_exit = graph.exit_node(first_map_fragment_entry)
            second_map_fragment_exit = graph.exit_node(second_map_fragment_entry)

            # Before we do anything we perform the renaming, i.e. we will rename the
            #  parameters of the second map such that they match the one of the first map.
            dace_mfhelper.rename_map_parameters(
                first_map=first_map_fragment_entry.map,
                second_map=second_map_fragment_entry.map,
                second_map_entry=second_map_fragment_entry,
                state=graph,
            )

            # Now we relocate all connectors from the second to the first map and remove
            #  the respective node of the second map.
            for to_node, from_node in [
                (first_map_fragment_entry, second_map_fragment_entry),
                (first_map_fragment_exit, second_map_fragment_exit),
            ]:
                dace_mfhelper.relocate_nodes(
                    from_node=from_node,
                    to_node=to_node,
                    state=graph,
                    sdfg=sdfg,
                    scope_dict=scope_dict,
                    never_consolidate_edges=self.never_consolidate_edges,
                    consolidate_edges_only_if_not_extending=self.consolidate_edges_only_if_not_extending,
                )
                # The relocate function does not remove the node, so we must do it.
                graph.remove_node(from_node)


@dace_properties.make_properties
class VerticalSplitMapRange(SplitMapRange):
    """
    Identify overlapping range between serial maps, and split the range in order
    to promote serial map fusion. In contrast to `HorizontalSplitMapRange` this
    class is not able to automatically fuse.
    """

    # Pattern Matching
    first_map_exit = dace_transformation.PatternNode(dace_nodes.MapExit)
    access_node = dace_transformation.PatternNode(dace_nodes.AccessNode)
    second_map_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def expressions(cls) -> Any:
        return [
            dace.sdfg.utils.node_path_graph(
                cls.first_map_exit, cls.access_node, cls.second_map_entry
            ),
        ]

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        """Check non overlapping range in the first and second map."""
        assert self.expr_index == expr_index
        first_map = self.first_map_exit.map
        first_map_entry: dace_nodes.MapEntry = graph.entry_node(self.first_map_exit)
        second_map = self.second_map_entry.map

        # Test if the map is in the right scope.
        map_scope: Union[dace_nodes.Node, None] = graph.scope_dict()[first_map_entry]
        if self.only_toplevel_maps and (map_scope is not None):
            return False

        if not self.access_node.desc(graph).transient:
            return False

        splitted_range = gtx_mfutils.split_overlapping_map_range(first_map, second_map)
        if splitted_range is None:
            return False

        # TODO(phimuell): Implement a check that the array is single use data?

        # Avoid spliting maps for fusion that if the second map has a nested map which is a neighbor reduction which
        # accesses the data that got generated by the first map. This is due to the fact that because of the neighbor
        # reduction, the second map might access a wider range of data that its range. This leads to a read-after-write
        # dependency between the first and second map.
        second_map_subgraph = graph.scope_subgraph(
            self.second_map_entry, include_entry=False, include_exit=False
        )
        for second_map_node in second_map_subgraph.nodes():
            if isinstance(second_map_node, dace_nodes.MapEntry):
                nested_map_input_edges_data = [
                    edge.data.data for edge in second_map_subgraph.in_edges(second_map_node)
                ]
                if self.access_node.data in nested_map_input_edges_data and any(
                    gtx_dace_utils.is_connectivity_identifier(data)
                    for data in nested_map_input_edges_data
                ):
                    return False

        # TODO: Ensure that the fusion can be performed.

        return True

    def apply(self, graph: Union[dace.SDFGState, dace.SDFG], sdfg: dace.SDFG) -> None:
        """Split the map range in order to obtain an overlapping range between the first and second map."""

        first_map_entry: dace_nodes.MapEntry = graph.entry_node(self.first_map_exit)
        first_map_exit: dace_nodes.MapExit = self.first_map_exit
        second_map_entry: dace_nodes.MapEntry = self.second_map_entry
        second_map_exit: dace_nodes.MapExit = graph.exit_node(self.second_map_entry)

        self.split_maps(
            graph, sdfg, first_map_entry, first_map_exit, second_map_entry, second_map_exit
        )
